"""CRC-16 ARC calculator with optional Rust acceleration.

Uses fastcrc (Rust extension) when available, falls back to a pre-computed
256-entry lookup table implementation.

CRC-16 ARC params: width=16, poly=0x8005, init=0, refIn=True, refOut=True, xorOut=0
"""

def make_crc16_arc():
    """Return a function: crc16_arc(data: bytes) -> int."""
    try:
        from fastcrc import crc16
        _arc = crc16.arc
        return lambda data: _arc(bytes(data))
    except ImportError:
        pass

    import warnings
    warnings.warn("fastcrc is not available, performance will be degraded.")

    table = [0] * 256
    for i in range(256):
        crc = i
        for _ in range(8):
            if crc & 1:
                crc = (crc >> 1) ^ 0xA001  # 0x8005 bit-reversed
            else:
                crc >>= 1
        table[i] = crc

    def crc16_arc(data):
        crc = 0
        for byte in data:
            crc = table[(crc ^ byte) & 0xFF] ^ (crc >> 8)
        return crc

    return crc16_arc



import sys
import json 
import glob
from collections import OrderedDict, defaultdict
import re
from pathlib import Path

def extract_module_names(file_path):
    """Extract module names from a Verilog file, preserving order."""
    with open(file_path, 'r', encoding='utf-8') as f:
        content = f.read()
    
    module_pattern = r'module\s+([A-Za-z0-9_]+)\s*[\(;]'
    modules = []
    for match in re.finditer(module_pattern, content):
        module_name = match.group(1)
        if module_name not in modules:
            modules.append(module_name)
    
    return modules

def analyze_used_primitives(json_pattern):
    """Extract primitives used in example designs."""
    primitive_counts = defaultdict(int)
    
    for filename in glob.glob(json_pattern):
        with open(filename) as f:
            design = json.load(f)
            
        for module in design['modules'].values():
            for cell in module.get('cells', {}).values():
                primitive_counts[cell['type']] += 1
    
    return set(primitive_counts.keys())

def escape_markdown(text):
    """Escape underscores for markdown."""
    return text.replace('_', '\\_')

def generate_markdown_table(files_modules, column_names, apicula_primitives):
    """Generate a markdown table including Apicula usage data."""
    all_modules = []
    for modules in files_modules.values():
        for module in modules:
            if module not in all_modules:
                all_modules.append(module)
    
    headers = ['Primitive'] + list(column_names.values()) + ['Apicula']
    header_row = '| ' + ' | '.join(headers) + ' |'
    separator_row = '| ' + ' | '.join(['---' for _ in headers]) + ' |'
    
    rows = []
    for module in all_modules:
        row = [f'[[{escape_markdown(module)}]]']
        for file in files_modules.keys():
            row.append('✓' if module in files_modules[file] else '')
        row.append('✓' if module in apicula_primitives else '')
        rows.append('| ' + ' | '.join(row) + ' |')
    
    return '\n'.join([header_row, separator_row] + rows)

def main():
    if len(sys.argv) < 3 or len(sys.argv) % 2 != 1:
        print("Usage: script.py <name1> <file1> <name2> <file2> ...")
        sys.exit(1)
    
    files_modules = OrderedDict()
    column_names = {}
    
    for i in range(1, len(sys.argv), 2):
        column_name = sys.argv[i]
        file_path = sys.argv[i + 1]
        files_modules[file_path] = extract_module_names(file_path)
        column_names[file_path] = column_name
    
    # Get primitives used in examples
    apicula_primitives = analyze_used_primitives('examples/himbaechel/*-synth.json')
    
    table = generate_markdown_table(files_modules, column_names, apicula_primitives)
    print(table)

if __name__ == "__main__":
    main()



import sys
import os
import re
import random
from itertools import chain, count
import argparse
import importlib.resources
from apycula import codegen
from apycula import chipdb
from apycula.chipdb import load_chipdb
from apycula import attrids
from apycula.bslib import read_bitstream, display

_device = ""
_pinout = ""
_packages = {
        'GW1N-1' : 'LQFP144', 'GW1NZ-1' : 'QFN48', 'GW1N-4' : 'PBGA256', 'GW1N-9C' : 'UBGA332',
        'GW1N-9' : 'PBGA256', 'GW1NS-4' : 'QFN48P', 'GW1NS-2' : 'LQFP144', 'GW2A-18': 'PBGA256',
        'GW2A-18C' : 'PBGA256S', 'GW5A-25A' : 'MBGA121N', 'GW5AST-138C' : 'PBGA484A'
}

def print_sorted_dict(start, d):
    print(start, end='{')
    for i in sorted(d):
        print(f'{i}:{d[i]}, ', end='')
    print('}')

# bank iostandards
# XXX default io standard may be board-dependent!
_banks = {'0': "LVCMOS18", '1': "LVCMOS18", '2': "LVCMOS18", '3': "LVCMOS18"}

# bank fuse tables. They are created here from the standard 'longval' because for
# banks the key of these tables starts with the bank number and unpack is not
# called so often that one can make 'right' tables on the fly.
_bank_fuse_tables = {}

# for a given mode returns a mask of zero bits
def zero_bits(mode, all_modes):
    res = set()
    for m, m_rec in all_modes.items():
        if m == mode:
            continue
        res.update(m_rec.decode_bits)
        for flag in m_rec.flags.values():
            res.update(flag.mask)
    m_mask = set()
    for flag in all_modes[mode].flags.values():
        m_mask.update(flag.mask)
    return res.difference(all_modes[mode].decode_bits).difference(m_mask)

# If the length of the bit pattern is equal, start the comparison with IOBUF
def _io_mode_sort_func(mode):
    l = len(mode[1].decode_bits) * 10
    if mode[0] == 'IOBUF':
        l += 2
    elif mode[0] == 'OBUF':
        l += 1
    return l

#
def get_attr_name(attrname_table, code, tableName):
    for name, cod in attrname_table.items():
        if cod == code:
            return name
    print(f'Unknown attr name for table: {tableName} code:{code}')
    return ''

# fix names and types of the PLL attributes
# { internal_name: external_name }
_pll_attrs = {
        'IDIV' :            'IDIV_SEL',
        'IDIVSEL' :         'DYN_IDIV_SEL',
        'FDIV' :            'FBDIV_SEL',
        'FDIVSEL' :         'DYN_FBDIV_SEL',
        'ODIV' :            'ODIV_SEL',
        'ODIVSEL' :         'DYN_ODIV_SEL',
        'PHASE' :           'PSDA_SEL',
        'DUTY' :            'DUTYDA_SEL',
        'DPSEL' :           'DYN_DA_EN',

        'OPDLY' :           'CLKOUT_DLY_STEP',

        'OSDLY' :           'CLKOUTP_DLY_STEP',
        'SDIV' :            'DYN_SDIV_SEL',

        'CLKOUTDIVSEL' :    'CLKOUTD_SRC',
        'CLKOUTDIV3SEL' :   'CLKOUTD3_SRC',
        'BYPCK' :           'CLKOUT_BYPASS',
        'BYPCKPS' :         'CLKOUTP_BYPASS',
        'BYPCKDIV' :        'CLKOUTD_BYPASS',
        }

_pll_vals = {
        'DYN' :             'true',
        'CLKOUTPS' :        'CLKOUTP',
        'BYPASS' :          'true',
        }
def pll_attrs_refine(in_attrs):
    res = set()
    for attr, val in in_attrs.items():
        #print(attr, val)
        if attr not in _pll_attrs.keys():
            if attr in ['INSEL', 'FBSEL', 'PWDEN', 'RSTEN', 'CLKOUTDIV3', 'CLKOUTPS']:
                res.add(f'{attr}="{[ name for name, vl in attrids.pll_attrvals.items() if vl == val ][0]}"')
            continue
        attr = _pll_attrs[attr]
        if attr in ['CLKOUTP_DLY_STEP', 'CLKOUT_DLY_STEP']:
            new_val = val / 50
        elif attr in ['PSDA_SEL', 'DUTYDA_SEL']:
            new_val = f'"{val:04b}"'
        elif attr in ['IDIV_SEL', 'FBDIV_SEL']:
            new_val = val - 1
        elif attr in ['DYN_SDIV_SEL', 'ODIV_SEL']:
            new_val = val
        else:
            attrvals = [ name for name, vl in attrids.pll_attrvals.items() if vl == val ]
            if not attrvals:
                raise Exception(f"PLL no {attr} = {val}")
            if attrvals[0] in _pll_vals.keys():
                new_val = _pll_vals[attrvals[0]]
            new_val = f'"{new_val}"'
        res.add(f'{attr}={new_val}')
    return res

_osc_attrs = {
        'MCLKCIB': 'FREQ_DIV',
        'OSCREG': 'REGULATOR_EN'
}

def osc_attrs_refine(in_attrs):
    res = set()
    for attr, val in in_attrs.items():
        if attr not in _osc_attrs.keys():
            continue
        attr = _osc_attrs[attr]
        if attr == 'FREQ_DIV':
            new_val = val
        else:
            attrvals = [ name for name, vl in osc_attrvals.items() if vl == val ]
            if attrvals[0] in osc_attrvals.keys():
                new_val = attrvals[0]
            new_val = f'"{new_val}"'
        res.add(f'{attr}={new_val}')
    if 'MCLKCIB' not in in_attrs.keys() and 'MCLKCIB_EN' in in_attrs.keys():
        res.add('FREQ_DIV=128')
    return res

# {(REGSET, LSRONMUX, CLKMUX_CLK, SRMODE) : dff_type}
_dff_types = {
   ('RESET', '',       'SIG', '') :      'DFF',
   ('RESET', '',       'INV', '') :      'DFFN',
   ('RESET', 'LSRMUX', 'SIG', 'ASYNC') : 'DFFC',
   ('RESET', 'LSRMUX', 'INV', 'ASYNC') : 'DFFNC',
   ('RESET', 'LSRMUX', 'SIG', '') :      'DFFR',
   ('RESET', 'LSRMUX', 'INV', '') :      'DFFNR',
   ('SET',   'LSRMUX', 'SIG', 'ASYNC') : 'DFFP',
   ('SET',   'LSRMUX', 'INV', 'ASYNC') : 'DFFNP',
   ('SET',   'LSRMUX', 'SIG', '') :      'DFFS',
   ('SET',   'LSRMUX', 'INV', '') :      'DFFNS',
}

# {(REGSET, LSRONMUX, CLKMUX_CLK, SRMODE) : latch_type}
_latch_types = {
   ('RESET', '',       'SIG', '') :      'DL',
   ('RESET', '',       'INV', '') :      'DLN',
   ('RESET', 'LSRMUX', 'SIG', 'ASYNC') : 'DLC',
   ('RESET', 'LSRMUX', 'INV', 'ASYNC') : 'DLNC',
   ('SET',   'LSRMUX', 'SIG', 'ASYNC') : 'DLP',
   ('SET',   'LSRMUX', 'INV', 'ASYNC') : 'DLNP',
}

def get_latch_type(dff_idx, in_attrs):
    def get_attrval_name(val):
        for nam, vl in attrids.cls_attrvals.items():
            if vl == val:
                return nam
        return None

    attrs = {}
    if 'LSRONMUX' in in_attrs.keys():
        attrs['LSRONMUX'] = get_attrval_name(in_attrs['LSRONMUX'])
    else:
        attrs['LSRONMUX'] = ''
    if 'CLKMUX_CLK' in in_attrs.keys():
        attrs['CLKMUX_CLK'] = get_attrval_name(in_attrs['CLKMUX_CLK'])
    else:
        attrs['CLKMUX_CLK'] = 'SIG'
    if 'SRMODE' in in_attrs.keys():
        attrs['SRMODE'] = get_attrval_name(in_attrs['SRMODE'])
    else:
        attrs['SRMODE'] = ''
    if f'REG{dff_idx % 2}_REGSET' in in_attrs.keys():
        attrs['REGSET'] = get_attrval_name(in_attrs[f'REG{dff_idx % 2}_REGSET'])
    else:
        attrs['REGSET'] = 'SET'
    return _latch_types.get((attrs['REGSET'], attrs['LSRONMUX'], attrs['CLKMUX_CLK'], attrs['SRMODE']))

def get_dff_type(dff_idx, in_attrs):
    def get_attrval_name(val):
        for nam, vl in attrids.cls_attrvals.items():
            if vl == val:
                return nam
        return None

    attrs = {}
    if 'LSRONMUX' in in_attrs.keys():
        attrs['LSRONMUX'] = get_attrval_name(in_attrs['LSRONMUX'])
    else:
        attrs['LSRONMUX'] = ''
    if 'CLKMUX_CLK' in in_attrs.keys():
        attrs['CLKMUX_CLK'] = get_attrval_name(in_attrs['CLKMUX_CLK'])
    else:
        attrs['CLKMUX_CLK'] = 'SIG'
    if 'SRMODE' in in_attrs.keys():
        attrs['SRMODE'] = get_attrval_name(in_attrs['SRMODE'])
    else:
        attrs['SRMODE'] = ''
    if f'REG{dff_idx % 2}_REGSET' in in_attrs.keys():
        attrs['REGSET'] = get_attrval_name(in_attrs[f'REG{dff_idx % 2}_REGSET'])
    else:
        attrs['REGSET'] = 'SET'
    return _dff_types.get((attrs['REGSET'], attrs['LSRONMUX'], attrs['CLKMUX_CLK'], attrs['SRMODE']))

# parse attributes and values use 'logicinfo' table
# returns {attr: value}
# attribute names are decoded with the attribute table, but the values are returned in raw form
def parse_attrvals(tile, logicinfo_table, fuse_table, attrname_table, tableName):
    def is_neg_key(key):
        for k in key:
            if k < 0:
                return True
        return False

    def is_pos_key(key):
        return not is_neg_key(key)

    def get_positive(av):
        return {a for a in av if a > 0}

    def get_negative(av):
        return {abs(a) for a in av if a < 0}

    res = {}
    set_mask = set()
    zero_mask = set()
    # collect masks
    for av, bits in fuse_table.items():
        if is_neg_key(av):
            zero_mask.update(bits)
        else:
            set_mask.update(bits)
    set_bits = {(row, col) for row, col in set_mask if tile[row][col] == 1}
    neg_bits = {(row, col) for row, col in zero_mask if tile[row][col] == 1}

    # find candidates from fuse table
    # the set bits are more unique
    attrvals = set()
    cnd = {av: bits for av, bits in fuse_table.items() if is_pos_key(av) and bits.issubset(set_bits)}
    for av, bits in cnd.items():
        keep = True
        for bt in cnd.values():
            if bits != bt and bits.issubset(bt):
                keep = False
                break
        if keep:
            clean_av = get_positive(av)
            attrvals.update(clean_av) # set attributes
            for idx in clean_av:
                attr, val = logicinfo_table[idx]
                res[get_attr_name(attrname_table, attr, tableName)] = val

    # records with a negative keys and used fuses
    neg_attrvals = set()
    ignore_attrs = set()
    cnd = {av: bits for av, bits in fuse_table.items() if is_neg_key(av) and bits.issubset(neg_bits)}
    for av, bits in cnd.items():
        keep = True
        for bt in cnd.values():
            if bits != bt and bits.issubset(bt):
                keep = False
                break
            for idx in av:
                attr, _ = logicinfo_table[abs(idx)]
                if attr in res.keys():
                    keep = False
                    break
        if keep:
            neg_attrvals.update(get_positive(av))
            ignore_attrs.update(get_negative(av))

    for idx in neg_attrvals:
        attr, val = logicinfo_table[idx]
        res[get_attr_name(attrname_table, attr, tableName)] = val

    # records with a negative keys and unused fuses
    cnd = {av for av, bits in fuse_table.items() if is_neg_key(av) and not bits.issubset(neg_bits)}
    for av in cnd:
        keep = True
        for idx in get_negative(av):
            if idx in ignore_attrs or not get_positive(av).issubset(attrvals):
                keep = False
                break
        if keep:
            for idx in get_negative(av):
                attr, val = logicinfo_table[idx]
                res[get_attr_name(attrname_table, attr, tableName)] = val
    return res

# { (row, col, type) : idx}
# type 'A'| 'B'
_pll_cells = {}

# returns the A cell of the PLL
def get_pll_A(db, row, col, typ):
    if typ == 'B':
        if _device in {"GW1N-9C", "GW1N-9"}:
            if col > 28:
                col = db.cols - 1
            else:
                col = 0
        else:
            col -= 1
    return row, col, 'A'

_iologic_mode = {
        'MODDRX2':  'OSER4',  'ODDRX2': 'OSER4',
        'MODDRX21': 'OSER4',  'ODDRX2': 'OSER4',
        'MODDRX4':  'OSER8',  'ODDRX4': 'OSER8',
        'MODDRX5':  'OSER10', 'ODDRX5': 'OSER10',
        'VIDEORX':  'OVIDEO', 'ODDRX8': 'OSER16',
        'MIDDRX2':  'IDES4',  'IDDRX2': 'IDES4',
        'MIDDRX4':  'IDES8',  'IDDRX4': 'IDES8',
        'MIDDRX5':  'IDES10', 'IDDRX5': 'IDES10',
        'IDDRX8':   'IDES16',
        }

# BSRAM has 3 cells: BSRAM, BSRAM0 and BSRAM1
# { (row, col) : idx }
_bsram_cells = {}
def get_bsram_main_cell(db, row, col, typ):
    if typ[-4:] == '_AUX':
        col -= 1
        if 'BSRAM_AUX' in db[row, col].bels:
            col -= 1
    return row, col

# The DSP has 9 cells: the main one and a group of auxiliary ones.
def get_dsp_main_cell(db, row, col, typ):
    if typ[-6:-2] == '_AUX':
        if _device in {'GW5A-25A'}:
            col = 1 + (col - 1) // 3
        else:
            col = 1 + (col - 1) // 9
    return row, col

# noiostd --- this is the case when the function is called
# with iostd by default, e.g. from the clock fuzzer
# With normal gowin_unpack io standard is determined first and it is known.
# (bels, pips, clock_pips)
def parse_tile_(db, row, col, tile, bm=None, default=True, noiostd = True):
    if not _bank_fuse_tables:
        # create bank fuse table
        for ttyp in db.longval.keys():
            if 'BANK' in db.longval[ttyp].keys():
                for key, val in db.longval[ttyp]['BANK'].items():
                    _bank_fuse_tables.setdefault(ttyp, {}).setdefault(f'BANK{key[0]}', {})[key[1:]] = val

    # TLVDS takes two BUF bels, so skip the B bels.
    skip_bels = set()
    #print((row, col))
    tiledata = db[row, col]
    #if 'HCLK' in db.shortval[tiledata.ttyp].keys():
    #    attrvals =parse_attrvals(tile, db.logicinfo['HCLK'], db.shortval[tiledata.ttyp]['HCLK'], attrids.hclk_attrids)
    #    if attrvals:
    #        print(row, col, attrvals)

    #if tiledata.ttyp in db.longfuses:
    #    if 'DLLDEL0' in db.longfuses[tiledata.ttyp].keys():
    #        attrvals =parse_attrvals(tile, db.logicinfo['DLLDLY'], db.longfuses[tiledata.ttyp]['DLLDEL0'], attrids.dlldly_attrids)
    #        if attrvals:
    #            print(row, col, attrvals)
    #    if 'DLLDEL1' in db.longfuses[tiledata.ttyp].keys():
    #        attrvals =parse_attrvals(tile, db.logicinfo['DLLDLY'], db.longfuses[tiledata.ttyp]['DLLDEL1'], attrids.dlldly_attrids)
    #        if attrvals:
    #            print(row, col, attrvals)

    clock_pips = {}
    bels = {}
    for name, bel in tiledata.bels.items():
        if name.startswith("ADC"):
            attrvals = parse_attrvals(tile, db.rev_logicinfo('ADC'), db.shortval[tiledata.ttyp]['ADC'], attrids.adc_attrids, "ADC")
            #print(row, col, name, tiledata.ttyp, attrvals)
        if name.startswith("RPLL"):
            idx = _pll_cells.setdefault(get_pll_A(db, row, col, name[4]), len(_pll_cells))
            modes = { f'DEVICE="{_device}"' }
            if 'PLL' in db.shortval[tiledata.ttyp].keys():
                attrvals = pll_attrs_refine(parse_attrvals(tile, db.rev_logicinfo('PLL'), db.shortval[tiledata.ttyp]['PLL'], attrids.pll_attrids, "PLL"))
                for attrval in attrvals:
                    modes.add(attrval)
            if modes:
                bels[f'{name}{idx}'] = modes
            continue
        if name == "PLLVR":
            idx = _pll_cells.setdefault(get_pll_A(db, row, col, 'A'), len(_pll_cells))
            attrvals = pll_attrs_refine(parse_attrvals(tile, db.rev_logicinfo('PLL'), db.shortval[tiledata.ttyp]['PLL'], attrids.pll_attrids, "PLL"))
            modes = { f'DEVICE="{_device}"' }
            for attrval in attrvals:
                modes.add(attrval)
            if modes:
                bels[f'{name}{idx}'] = modes
            continue
        if name.startswith("OSC"):
            attrvals = osc_attrs_refine(parse_attrvals(tile, db.rev_logicinfo('OSC'), db.shortval[tiledata.ttyp]['OSC'], attrids.osc_attrids, "OSC"))
            modes = set()
            for attrval in attrvals:
                modes.add(attrval)
            if modes:
                bels[name] = modes
            continue
        if name.startswith("BSRAM"):
            # disabled BSRAM cells have no fuse tables
            if 'BSRAM_SP' not in db.shortval[tiledata.ttyp]:
                continue
            idx = _bsram_cells.setdefault(get_bsram_main_cell(db, row, col, name), len(_bsram_cells))
            #print(row, col, name, idx, tiledata.ttyp)
            attrvals = parse_attrvals(tile, db.rev_logicinfo('BSRAM'), db.shortval[tiledata.ttyp]['BSRAM_SP'], attrids.bsram_attrids, "BSRAM")
            if not attrvals:
                continue
            #print(row, col, name, idx, tiledata.ttyp, attrvals)
            bels[f'{name}{idx}'] = {}
            continue
        if name.startswith("ALU54D"):
            continue
        if name.startswith("DSP") or name.startswith("DSP_AUX"):
            modes = set()
            idx = name[-1]
            #print(row, col, name, idx, tiledata.ttyp)
            if name.startswith("DSP_AUX"):
                row, col = get_dsp_main_cell(db, row, col, name)

            if f'DSP{idx}' in db.shortval[tiledata.ttyp]:
                attrvals = parse_attrvals(tile, db.rev_logicinfo('DSP'), db.shortval[tiledata.ttyp][f'DSP{idx}'], attrids.dsp_attrids, "DSP")
            elif f'5A_DSP' in db.shortval[tiledata.ttyp]:
                attrvals = parse_attrvals(tile, db.rev_logicinfo('5A_DSP'), db.shortval[tiledata.ttyp][f'5A_DSP'], attrids.dsp_5a_attrids, "5A_DSP")
            else:
                attrvals = None
            if attrvals:
                #print_sorted_dict(f'{row}, {col}, {name}, {idx}, {tiledata.ttyp} - ', attrvals)
                for attrval in attrvals:
                    modes.add(attrval)
            if modes and not name.startswith("DSP_AUX"):
                bels[f'{name}{idx}'] = modes
            continue
        if name.startswith("IOLOGIC"):
            idx = name[-1]
            attrvals = parse_attrvals(tile, db.rev_logicinfo('IOLOGIC'), db.shortval[tiledata.ttyp][f'IOLOGIC{idx}'], attrids.iologic_attrids, "IOLOGIC")
            if not attrvals:
                continue
            #print_sorted_dict(f'{row}, {col}, {name}, {idx}, {tiledata.ttyp} - ', attrvals)
            # additional IOLOGIC components
            # XXX delays and FFs in IO
            # main component
            if 'OUTMODE' in attrvals.keys():
                # XXX skip oddr
                if attrvals['OUTMODE'] in {attrids.iologic_attrvals['MODDRX1'], attrids.iologic_attrvals['ODDRX1']}:
                    if 'LSROMUX_0' in attrvals.keys():
                        bels.setdefault(name, set()).add(f"MODE=ODDRC")
                    else:
                        bels.setdefault(name, set()).add(f"MODE=ODDR")
                    continue
                # skip aux cells
                if attrvals['OUTMODE'] == attrids.iologic_attrvals['DDRENABLE']:
                    continue
                if attrids.iologic_num2val[attrvals['OUTMODE']] in _iologic_mode.keys():
                    bels.setdefault(name, set()).add(f"MODE={_iologic_mode[attrids.iologic_num2val[attrvals['OUTMODE']]]}")
            elif 'INMODE' in attrvals.keys():
                if attrvals['INMODE'] in {attrids.iologic_attrvals['MIDDRX1'], attrids.iologic_attrvals['IDDRX1']}:
                    if 'LSRIMUX_0' in attrvals.keys():
                        bels.setdefault(name, set()).add(f"MODE=IDDRC")
                    else:
                        bels.setdefault(name, set()).add(f"MODE=IDDR")
                    continue
                # skip aux cells
                if attrvals['INMODE'] == attrids.iologic_attrvals['DDRENABLE']:
                    continue
                if attrids.iologic_num2val[attrvals['INMODE']] in _iologic_mode.keys():
                    in_mode = _iologic_mode[attrids.iologic_num2val[attrvals['INMODE']]]
                    if in_mode == 'OVIDEO':
                        in_mode = 'IVIDEO'
                    bels.setdefault(name, set()).add(f"MODE={in_mode}")
            else:
                continue
            if 'CLKODDRMUX_ECLK' in attrvals.keys():
                bels.setdefault(name, set()).add(f"CLKODDRMUX_ECLK={attrids.iologic_num2val[attrvals['CLKODDRMUX_ECLK']]}")
        if name.startswith("DFF"):
            idx = int(name[3])
            attrvals = parse_attrvals(tile, db.rev_logicinfo('SLICE'), db.shortval[tiledata.ttyp][f'CLS{idx // 2}'], attrids.cls_attrids, "CLS")
            #print('parse', row, col, attrvals)
            # skip ALU and unsupported modes
            if attrvals.get('MODE') == attrids.cls_attrvals['SSRAM']:
                continue
            if attrvals.get('REGMODE') == attrids.cls_attrvals['LATCH']:
                latch_type = get_latch_type(idx, attrvals)
                if latch_type:
                    bels[f'{name}'] = {latch_type}
                continue
            dff_type = get_dff_type(idx, attrvals)
            if dff_type:
                bels[f'{name}'] = {dff_type}
            if f'REG{idx % 2}_SD' in attrvals:
                bels[f'{name}'].update({'SD'})
            continue
        if name.startswith("IOB"):
            idx = name[-1]
            io_row, io_col = row, col
            io_tile = tile
            io_ttyp = tiledata.ttyp

            if idx == 'B' and tiledata.bels[name].fuse_cell_offset:
                io_row += tiledata.bels[name].fuse_cell_offset[0]
                io_col += tiledata.bels[name].fuse_cell_offset[1]
                io_tiledata = db[io_row, io_col]
                io_tile = bm[io_row, io_col]
                io_ttyp = io_tiledata.ttyp
            # XXX
            if idx == 'B' and 'IOBB' not in db.longval[io_ttyp]:
                continue
            attrvals = parse_attrvals(io_tile, db.rev_logicinfo('IOB'), db.longval[io_ttyp][f'IOB{idx}'], attrids.iob_attrids, "IOB")
            #print(name, io_row, io_col, sorted(attrvals.items()))
            try: # we can ask for invalid pin here because the IOBs share some stuff
                bank = chipdb.loc2bank(db, io_row, io_col)
            except KeyError:
                bank = None
            if attrvals:
                mode = 'IBUF'
                if attrvals.get('PERSISTENT', None) == attrids.iob_attrvals['OFF']:
                    mode = 'IOBUF'
                elif 'ODMUX' in attrvals or 'ODMUX_1' in attrvals:
                    mode = 'OBUF'
                # Z-1 row 6
                if _device in {'GW1NZ-1', 'GW1N-1'} and row == 5:
                    mode = 'IOBUF'
                if 'LVDS_OUT' in attrvals:
                    if mode == 'IOBUF':
                        mode = 'TBUF'
                    if 'MIPI' in attrvals:
                        mode = 'MIPI_OBUF'
                    else:
                        mode = f'TLVDS_{mode}'
                    # skip B bel
                    skip_bels.update({name[:-1] + 'B'})
                elif 'OD' in attrvals:
                    mode = 'I3C_IOBUF'
                elif idx == 'B' and 'DRIVE' not in attrvals and 'IO_TYPE' in attrvals:
                    mode = f'ELVDS_{mode}'
                    # skip B bel
                    skip_bels.update({name})
                elif 'LPRX_A1' in attrvals:
                    mode = f'MIPI_IBUF'
                    # skip B bel
                    skip_bels.update({name[:-1] + 'B'})
                elif 'IOBUF_MIPI_LP' in attrvals:
                    mode = f'ELVDS_{mode}'
                    # skip B bel
                    skip_bels.update({name[:-1] + 'B'})

                bels.setdefault(name, set()).add(mode)
        if name.startswith("BANK"):
            attrvals = parse_attrvals(tile, db.rev_logicinfo('IOB'), _bank_fuse_tables[tiledata.ttyp][name], attrids.iob_attrids, "IOB")
            #print(name, row, col, attrvals)

            for a, v in attrvals.items():
                bels.setdefault(name, set()).add(f'{a}={attrids.iob_num2val.get(v, str(v))}')
        if name.startswith("ALU"):
            idx = int(name[3])
            attrvals = parse_attrvals(tile, db.rev_logicinfo('SLICE'), db.shortval[tiledata.ttyp][f'CLS{idx // 2}'], attrids.cls_attrids, "CLS")
            # skip ALU and unsupported modes
            if attrvals.get('MODE') != attrids.cls_attrvals['ALU']:
                continue
            bels[name] = {"C2L"}
            mode_bits = {(row, col)
                         for row, col in bel.mode_bits
                         if tile[row][col] == 1}
            for mode, bits in bel.modes.items():
                if bits == mode_bits and (default or bits):
                    bels[name] = {mode}
        else:
            mode_bits = {(row, col)
                         for row, col in bel.mode_bits
                         if tile[row][col] == 1}
            #print(name, sorted(bel.mode_bits))
            #print("read mode:", sorted(mode_bits))
            for mode, bits in bel.modes.items():
                #print(mode, sorted(bits))
                if bits == mode_bits and (default or bits):
                    bels.setdefault(name, set()).add(mode)
        # simple flags
        for flag, bits in bel.flags.items():
            used_bits = {tile[row][col] for row, col in bits}
            if all(used_bits):
                if name == "RAM16" and not name in bels:
                    continue
                bels.setdefault(name, set()).add(flag)
        # revert BUFS flags
        if name.startswith('BUFS'):
            flags = bels.get(name, set()) ^ {'R', 'L'}
            if flags:
                num = name[4:]
                half = 'T'
                if row != 0:
                    half = 'B'
                for qd in flags:
                    clock_pips[f'LWSPINE{half}{qd}{num}'] = f'LW{half}{num}'
        #print("flags:", sorted(bels.get(name, set())))

    pips = {}
    for dest, srcs in tiledata.pips.items():
        pip_bits = set().union(*srcs.values())
        used_bits = {(row, col)
                     for row, col in pip_bits
                     if tile[row][col] == 1}
        for src, bits in srcs.items():
            # optionally ignore the defautl set() state
            if bits == used_bits and (default or bits):
                pips[dest] = src

    for dest, srcs in tiledata.clock_pips.items():
        pip_bits = set().union(*srcs.values())
        used_bits = {(row, col)
                     for row, col in pip_bits
                     if tile[row][col] == 1}
        for src, bits in srcs.items():
            if bits == used_bits:
                clock_pips[dest] = src

    # elvds IO uses the B bel bits
    for name in skip_bels:
        bel_a_name = f'{name[:-1]}A'
        if bel_a_name not in bels:
            continue

        bel_a = bels[bel_a_name]
        if not bel_a.intersection({'ELVDS_IBUF', 'ELVDS_OBUF', 'ELVDS_IOBUF', 'ELVDS_TBUF',
                                   'TLVDS_IBUF', 'TLVDS_OBUF', 'TLVDS_IOBUF', 'TLVDS_TBUF', 'MIPI_IBUF', 'MIPI_OBUF'}):
            mode = bels[name].intersection({'ELVDS_IBUF', 'ELVDS_OBUF', 'ELVDS_IOBUF', 'ELVDS_TBUF'})
            if mode:
                old_mode = bel_a.intersection({'IBUF', 'OBUF', 'IOBUF', 'TBUF'})
                bel_a -= old_mode
                bel_a.update(mode)

    return {name: bel for name, bel in bels.items() if name not in skip_bels}, pips, clock_pips

dffmap = {
    "DFF": None,
    "DFFN": None,
    "DFFS": "SET",
    "DFFR": "RESET",
    "DFFP": "PRESET",
    "DFFC": "CLEAR",
    "DFFNS": "SET",
    "DFFNR": "RESET",
    "DFFNP": "PRESET",
    "DFFNC": "CLEAR",
}
latchmap = {
    "DL": None,
    "DLN": None,
    "DLC": "CLEAR",
    "DLNC": "CLEAR",
    "DLP": "PRESET",
    "DLNP": "PRESET",
}
iobmap = {
    "IBUF": {"wires": ["O"], "inputs": ["I"]},
    "OBUF": {"wires": ["I"], "outputs": ["O"]},
    "TBUF": {"wires": ["I", "OE"], "outputs": ["O"]},
    "IOBUF": {"wires": ["I", "O", "OE"], "inouts": ["IO"]},
    "TLVDS_OBUF": {"wires": ["I"], "outputs": ["O", "OB"]},
    "TLVDS_TBUF": {"wires": ["I", "OE"], "outputs": ["O", "OB"]},
    "TLVDS_IBUF": {"wires": ["O"], "inputs": ["I", "IB"]},
    "ELVDS_OBUF": {"wires": ["I"], "outputs": ["O", "OB"]},
    "ELVDS_TBUF": {"wires": ["I", "OE"], "outputs": ["O", "OB"]},
    "ELVDS_IBUF": {"wires": ["O"], "inputs": ["I", "IB"]},
    "ELVDS_IOBUF": {"wires": ["I", "O", "OE"], "inouts": ["IO", "IOB"]},
    "MIPI_IBUF": {"wires": ["I", "IB", "OEN", "OENB", "HSREN", "OH", "OL", "OB"], "inouts": ["IO", "IOB"]},
    "MIPI_OBUF": {"wires": ["I", "IB", "IL", "MODESEL"], "inouts": ["O", "OB"]},
    "I3C_IOBUF": {"wires": ["I", "O", "MODESEL"], "inouts": ["IO"]},
}

# OE -> OEN
def portname(n):
    if n == "OE":
        return "OEN"
    return n

def make_muxes(row, col, idx, db, mod):
    name = f"R{row}C{col}_MUX2_LUT50"
    if name in mod.primitives.keys():
        return

    # one MUX8
    if col < db.cols :
        name = f"R{row}C{col}_MUX2_LUT80"
        mux2 = codegen.Primitive("MUX2", name)
        mux2.portmap['I0'] = f"R{row}C{col + 1}_OF3"
        mux2.portmap['I1'] = f"R{row}C{col}_OF3"
        mux2.portmap['O']  = f"R{row}C{col}_OF7"
        mux2.portmap['S0'] = f"R{row}C{col}_SEL7"
        mod.wires.update(mux2.portmap.values())
        mod.primitives[name] = mux2

    # one MUX7
    name = f"R{row}C{col}_MUX2_LUT70"
    mux2 = codegen.Primitive("MUX2", name)
    mux2.portmap['I0'] = f"R{row}C{col}_OF5"
    mux2.portmap['I1'] = f"R{row}C{col}_OF1"
    mux2.portmap['O']  = f"R{row}C{col}_OF3"
    mux2.portmap['S0'] = f"R{row}C{col}_SEL3"
    mod.wires.update(mux2.portmap.values())
    mod.primitives[name] = mux2

    # two MUX6
    for i in range(2):
        name = f"R{row}C{col}_MUX2_LUT6{i}"
        mux2 = codegen.Primitive("MUX2", name)
        mux2.portmap['I0'] = f"R{row}C{col}_OF{i * 4 + 2}"
        mux2.portmap['I1'] = f"R{row}C{col}_OF{i * 4}"
        mux2.portmap['O']  = f"R{row}C{col}_OF{i * 4 + 1}"
        mux2.portmap['S0'] = f"R{row}C{col}_SEL{i * 4 + 1}"
        mod.wires.update(mux2.portmap.values())
        mod.primitives[name] = mux2

    # four MUX5
    for i in range(4):
        name = f"R{row}C{col}_MUX2_LUT5{i}"
        mux2 = codegen.Primitive("MUX2", name)
        mux2.portmap['I0'] = f"R{row}C{col}_F{i * 2}"
        mux2.portmap['I1'] = f"R{row}C{col}_F{i * 2  + 1}"
        mux2.portmap['O']  = f"R{row}C{col}_OF{i * 2}"
        mux2.portmap['S0'] = f"R{row}C{col}_SEL{i * 2}"
        mod.wires.update(mux2.portmap.values())
        mod.primitives[name] = mux2

_alu_re = re.compile(r"ALU(\w*)")
def removeLUTs(bels):
    bels_to_remove = []
    for bel in bels:
        match = _alu_re.match(bel)
        if match:
            bels_to_remove.append(f"LUT{match.group(1)}")
    for bel in bels_to_remove:
        bels.pop(bel, None)

def removeALUs(bels):
    bels_to_remove = []
    for bel in bels:
        match = _alu_re.match(bel)
        if match:
            bels_to_remove.append(match.group(0))
    for bel in bels_to_remove:
        bels.pop(bel, None)

def ram16_remove_bels(bels):
    bels_to_remove = []
    for bel in bels:
        if bel == "RAM16":
            bels_to_remove.extend(f"LUT{x}" for x in range(6))
            bels_to_remove.extend(f"DFF{x}" for x in range(4, 6))
    for bel in bels_to_remove:
        bels.pop(bel, None)

_iologic_bels = ['IOLOGICA', 'IOLOGICB', 'ODDRA', 'ODDRB']
def move_iologic(bels):
    res = []
    for iol_bel in _iologic_bels:
        if iol_bel in bels.keys():
            res.append((iol_bel, bels[iol_bel]))
    res += [(bel, flags) for bel, flags in bels.items() if bel not in _iologic_bels]
    return res

def disable_unused_pll_ports(pll):
    if 'DYN_DA_EN' not in pll.params:
        for n in range(0, 4):
            if f'PSDA{n}' in pll.portmap:
                del pll.portmap[f'PSDA{n}']
                del pll.portmap[f'DUTYDA{n}']
                del pll.portmap[f'FDLY{n}']
    if 'DYN_IDIV_SEL' not in pll.params:
        for n in range(0, 6):
            if f'IDSEL{n}' in pll.portmap:
                del pll.portmap[f'IDSEL{n}']
    if 'DYN_FBDIV_SEL' not in pll.params:
        for n in range(0, 6):
            if f'FBDSEL{n}' in pll.portmap:
                del pll.portmap[f'FBDSEL{n}']
    if 'DYN_ODIV_SEL' not in pll.params:
        for n in range(0, 6):
            if f'ODSEL{n}' in pll.portmap:
                del pll.portmap[f'ODSEL{n}']
    if 'PWDEN' in pll.params:
        if pll.params['PWDEN'] == 'DISABLE':
            if 'RESET_P' in pll.portmap:
                del pll.portmap['RESET_P']
        del pll.params['PWDEN']
    if 'RSTEN' in pll.params:
        if pll.params['RSTEN'] == 'DISABLE':
            if 'RESET' in pll.portmap:
                del pll.portmap['RESET']
        del pll.params['RSTEN']
    if 'CLKOUTDIV3' in pll.params:
        if pll.params['CLKOUTDIV3'] == 'DISABLE':
            if 'CLKOUTD3' in pll.portmap:
                del pll.portmap['CLKOUTD3']
        del pll.params['CLKOUTDIV3']
    if 'CLKOUTDIV' in pll.params:
        if pll.params['CLKOUTDIV'] == 'DISABLE':
            if 'CLKOUTD' in pll.portmap:
                del pll.portmap['CLKOUTD']
        del pll.params['CLKOUTDIV']
    if 'CLKOUTPS' in pll.params:
        if pll.params['CLKOUTPS'] == 'DISABLE':
            if 'CLKOUTP' in pll.portmap:
                del pll.portmap['CLKOUTP']
        del pll.params['CLKOUTPS']

_tbrlre = re.compile(r"IO([TBRL])(\d+)(\w)")
def tbrl2rc(db, loc):
    side, num, bel_idx = _tbrlre.match(loc).groups()
    if side == 'T':
        row = 0
        col = int(num) - 1
    elif side == 'B':
        row = db.rows - 1
        col = int(num) - 1
    elif side == 'L':
        row = int(num) - 1
        col = 0
    elif side == 'R':
        row = int(num) - 1
        col = db.cols - 1
    return (row, col, bel_idx)

def find_pll_in_pin(db, pll):
    locs = [loc for (loc, cfgs) in _pinout.values() if 'RPLL_T_IN' in cfgs or 'LRPLL_T_IN' in cfgs]
    if not locs:
        raise Exception(f"No [RL]PLL_T_IN pin in the current package")
    row, col, bel_idx = tbrl2rc(db, locs[0])
    wire = db[row, col].bels[f'IOB{bel_idx}'].portmap['O']
    pll.portmap['CLKIN'] = f'R{row + 1}C{col + 1}_{wire}'

def modify_pll_inputs(db, pll):
    if 'INSEL' in pll.params.keys():
        insel = pll.params['INSEL']
        if insel != 'CLKIN1':
            # pin
            if insel == 'CLKIN0':
                find_pll_in_pin(db, pll)
            else:
                if 'CLKIN' in pll.portmap:
                    del pll.portmap['CLKIN']
        del pll.params['INSEL']
    if 'FBSEL' in pll.params.keys():
        fbsel = pll.params['FBSEL']
        if fbsel == 'CLKFB3':
            # internal
            pll.params['CLKFB_SEL'] = '"internal"'
            if 'CLKFB' in pll.portmap:
                del pll.portmap['CLKFB']
        elif fbsel == 'CLKFB0':
            # external CLK2
            pll.params['CLKFB_SEL'] = '"external"'
        elif fbsel == 'CLKFB2':
            # external pin
            pll.params['CLKFB_SEL'] = '"external"'
            # XXX find pin
        del pll.params['FBSEL']

_iologic_ports = {
        'ODDR' :  {'D0': 'D0', 'D1': 'D1', 'Q0': 'Q0', 'Q1': 'Q1', 'CLK': 'CLK'},
        'ODDRC' : {'D0': 'D0', 'D1': 'D1', 'Q0': 'Q0', 'Q1': 'Q1', 'CLK': 'CLK', 'CLEAR': 'CLEAR'},
        'OSER4': {'D0': 'D0', 'D1': 'D1', 'D2': 'D2', 'D3': 'D3',
                  'Q0': 'Q0', 'Q1': 'Q1', 'RESET': 'RESET', 'TX0': 'TX0',
                  'TX1': 'TX1', 'PCLK': 'PCLK', 'FCLK': 'FCLK'},
        'OSER8': {'D0': 'D0', 'D1': 'D1', 'D2': 'D2', 'D3': 'D3',
                  'D4': 'D4', 'D5': 'D5', 'D6': 'D6', 'D7': 'D7',
                  'Q0': 'Q0', 'Q1': 'Q1', 'RESET': 'RESET', 'TX0': 'TX90',
                  'TX1': 'TX1', 'TX2': 'TX2', 'TX3': 'TX3',
                  'PCLK': 'PCLK', 'FCLK': 'FCLK'},
        'OVIDEO':{'D0': 'D0', 'D1': 'D1', 'D2': 'D2', 'D3': 'D3',
                  'D4': 'D4', 'D5': 'D5', 'D6': 'D6', 'Q': 'Q',
                  'RESET': 'RESET', 'PCLK': 'PCLK', 'FCLK': 'FCLK'},
        'OSER10': {'D0': 'D0', 'D1': 'D1', 'D2': 'D2', 'D3': 'D3',
                   'D4': 'D4', 'D5': 'D5', 'D6': 'D6', 'D7': 'D7', 'D8': 'D8', 'D9': 'D9',
                   'Q': 'Q', 'RESET': 'RESET', 'PCLK': 'PCLK', 'FCLK': 'FCLK'},
        'OSER16': {'D0': 'A0', 'D1': 'A1', 'D2': 'A2', 'D3': 'A3',
                   'D4': 'C1', 'D5': 'C0', 'D6': 'D1', 'D7': 'D0', 'D8': 'C3', 'D9': 'C2',
                   'D10': 'B4', 'D11': 'B5', 'D12': 'A0', 'D13': 'A1', 'D14': 'A2',
                   'D15': 'A3',},
        'IDDR' :  {'D': 'D', 'Q8': 'Q0', 'Q9': 'Q1', 'CLK': 'CLK'},
        'IDDRC' : {'D': 'D', 'Q8': 'Q0', 'Q9': 'Q1', 'CLK': 'CLK', 'CLEAR': 'CLEAR'},
        'IDES4':  {'D': 'D', 'Q6': 'Q0', 'Q7': 'Q1', 'Q8': 'Q2', 'Q9': 'Q3',
                   'RESET': 'RESET', 'CALIB': 'CALIB', 'PCLK': 'PCLK', 'FCLK': 'FCLK'},
        'IDES8': {'D': 'D', 'Q2': 'Q0', 'Q3': 'Q1', 'Q4': 'Q2', 'Q5': 'Q3', 'Q6': 'Q4',
                  'Q7': 'Q5', 'Q8': 'Q6', 'Q9': 'Q7',
                  'RESET': 'RESET', 'CALIB': 'CALIB', 'PCLK': 'PCLK', 'FCLK': 'FCLK'},
        'IVIDEO': {'D': 'D', 'Q3': 'Q0', 'Q4': 'Q1', 'Q5': 'Q2', 'Q6': 'Q3', 'Q7': 'Q4',
                   'Q8': 'Q5', 'Q9': 'Q6',
                   'RESET': 'RESET', 'CALIB': 'CALIB', 'PCLK': 'PCLK', 'FCLK': 'FCLK'},
        'IDES10': {'D': 'D', 'Q0': 'Q0', 'Q1': 'Q1', 'Q2': 'Q2', 'Q3': 'Q3', 'Q4': 'Q4',
                   'Q5': 'Q5', 'Q6': 'Q6', 'Q7': 'Q7', 'Q8': 'Q8', 'Q9': 'Q9',
                   'RESET': 'RESET', 'CALIB': 'CALIB', 'PCLK': 'PCLK', 'FCLK': 'FCLK'},
        'IDES16': {'Q0': 'F2', 'Q1': 'F3', 'Q2': 'F4', 'Q3': 'F5', 'Q4': 'Q0', 'Q5': 'Q1',
                   'Q6': 'Q2', 'Q7': 'Q3', 'Q8': 'Q4', 'Q9': 'Q5', 'Q10': 'F0',
                   'Q11': 'F1', 'Q12': 'F2', 'Q13': 'F3', 'Q14': 'F4', 'Q15': 'F5' },
}
def iologic_ports_by_type(typ, portmap):
    if typ not in {'IDES16', 'OSER16'}:
        return { (_iologic_ports[typ][port], wire) for port, wire in portmap.items() if port in _iologic_ports[typ].keys() }
    elif typ in {'OSER16', 'IDES16'}:
        ports = { (port, wire) for port, wire in _iologic_ports[typ].items()}
        ports.add(('RESET', portmap['RESET']))
        ports.add(('PCLK', portmap['PCLK']))
        ports.add(('FCLK', portmap['FCLK']))
        if typ == 'IDES16':
            ports.add(('CALIB', portmap['CALIB']))
            ports.add(('D', portmap['D']))
        else:
            ports.add(('Q', portmap['Q']))
        return ports

_sides = "AB"
def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
    # db is 0-based, floorplanner is 1-based
    row = dbrow+1
    col = dbcol+1

    for dest, src in chain(pips.items(), clock_pips.items()):
        srcg = chipdb.wire2global(row, col, db, src)
        destg = chipdb.wire2global(row, col, db, dest)
        mod.wires.update({srcg, destg})
        mod.assigns.append((destg, srcg))

    belre = re.compile(r"(IOB|LUT|DFF|BANK|CFG|ALU|RAM16|ODDR|OSC[ZFHWO]?|BUFS|RPLL[AB]|PLLVR|IOLOGIC|BSRAM|DSP)(\w*)")
    bels_items = move_iologic(bels)

    iologic_detected = set()
    disable_oddr = False
    for bel, flags in bels_items:
        typ, idx = belre.match(bel).groups()

        if typ == "LUT":
            val = 0xffff - sum(1<<f for f in flags)
            if val == 0:
                mod.assigns.append((f"R{row}C{col}_F{idx}", "VSS"))
            else:
                name = f"R{row}C{col}_LUT4_{idx}"
                lut = codegen.Primitive("LUT4", name)
                lut.params["INIT"] = f"16'h{val:04x}"
                lut.portmap['F'] = f"R{row}C{col}_F{idx}"
                lut.portmap['I0'] = f"R{row}C{col}_A{idx}"
                lut.portmap['I1'] = f"R{row}C{col}_B{idx}"
                lut.portmap['I2'] = f"R{row}C{col}_C{idx}"
                lut.portmap['I3'] = f"R{row}C{col}_D{idx}"
                mod.wires.update(lut.portmap.values())
                mod.primitives[name] = lut
                cst.cells[name] = (row, col, int(idx) // 2, _sides[int(idx) % 2])
            make_muxes(row, col, idx, db, mod)
        elif typ.startswith("IOLOGIC"):
            iologic_detected.add(idx)
            iol_mode = 'IVIDEO' #XXX
            disable_oddr = True
            eclk = 'HCLK0'
            iol_params = {}
            for paramval in flags:
                param, _, val = paramval.partition('=')
                if param == 'MODE':
                    iol_mode = val
                    if val == 'OSER4':
                        disable_oddr = False
                    continue
                if param == 'CLKODDRMUX_ECLK':
                    eclk == val
                    continue
                if param == 'CLKIDDRMUX_ECLK':
                    eclk == val
                    continue
                iol_params[param] = val
            name = f"R{row}C{col}_{iol_mode}_{idx}"
            iol = mod.primitives.setdefault(name, codegen.Primitive(iol_mode, name))
            iol.params.update(iol_params)
            iol_oser = iol_mode in {'ODDR', 'ODDRC', 'OSER4', 'OVIDEO', 'OSER8', 'OSER10', 'OSER16'}

            portmap = db[dbrow, dbcol].bels[bel].portmap
            for port, wname in iologic_ports_by_type(iol_mode, portmap):
                if iol_oser:
                    if port in {'Q', 'Q0', 'Q1'}:
                        if port == 'Q1':
                            iol.portmap[port] = f"R{row}C{col}_{portmap['TX0']}_IOL"
                        else:
                            iol.portmap[port] = f"R{row}C{col}_{portmap['D0']}_IOL"
                    elif port == 'FCLK':
                        wname = eclk
                        if eclk == 'HCLK0' and _device in {'GW1N-1'}:
                            wname = 'CLK2'
                        iol.portmap[port] = f"R{row}C{col}_{wname}"
                    else:
                        if iol_mode != 'OSER16' or port not in {'D12', 'D13', 'D14', 'D15'}:
                            iol.portmap[port] = f"R{row}C{col}_{wname}"
                        else:
                            if row == 1 or row == db.rows:
                                iol.portmap[port] = f"R{row}C{col + 1}_{wname}"
                            else:
                                iol.portmap[port] = f"R{row + 1}C{col}_{wname}"
                else: # IDES
                    if port in {'D'}:
                        iol.portmap[port] = f"R{row}C{col}_{portmap['D']}_IOL"
                    else:
                        if iol_mode != 'IDES16':
                            iol.portmap[port] = f"R{row}C{col}_{wname}"
                        else:
                            if port not in {'Q0', 'Q1', 'Q2', 'Q3'}:
                                iol.portmap[port] = f"R{row}C{col}_{wname}"
                            else:
                                if row == 1 or row == db.rows:
                                    iol.portmap[port] = f"R{row}C{col + 1}_{wname}"
                                else:
                                    iol.portmap[port] = f"R{row + 1}C{col}_{wname}"
                if port == 'FCLK':
                    wname = eclk
                    if eclk == 'HCLK0' and _device in {'GW1N-1'}:
                        wname = 'CLK2'
                    iol.portmap[port] = f"R{row}C{col}_{wname}"

        elif typ.startswith("RPLL"):
            name = f"PLL_{idx}"
            pll = mod.primitives.setdefault(name, codegen.Primitive("rPLL", name))
            for paramval in flags:
                param, _, val = paramval.partition('=')
                pll.params[param] = val
            portmap = db[dbrow, dbcol].bels[bel[:5]].portmap
            for port, wname in portmap.items():
                pll.portmap[port] = f"R{row}C{col}_{wname}"
        elif typ.startswith("PLLVR"):
            name = f"PLL_{idx}"
            pll = mod.primitives.setdefault(name, codegen.Primitive("PLLVR", name))
            for paramval in flags:
                param, _, val = paramval.partition('=')
                pll.params[param] = val
            portmap = db[dbrow, dbcol].bels[bel[:-1]].portmap
            for port, wname in portmap.items():
                pll.portmap[port] = f"R{row}C{col}_{wname}"
        elif typ.startswith("BSRAM"):
            #print(dbrow, dbcol, typ, bel, idx)
            if idx.startswith("_AUX"):
                continue
            bel_name = "BSRAM"
            name = f"BSRAM_{idx}"
            pll = mod.primitives.setdefault(name, codegen.Primitive("BSRAM", name))
            for paramval in flags:
                param, _, val = paramval.partition('=')
                pll.params[param] = val
            portmap = db[dbrow, dbcol].bels[bel_name].portmap
            for port, wname in portmap.items():
                pll.portmap[port] = f"R{row}C{col}_{wname}"
        elif typ == "ALU":
            #print(flags)
            kind, = flags # ALU only have one flag
            idx = int(idx)
            name = f"R{row}C{col}_ALU_{idx}"
            if kind == 'hadder':
                kind = '0'
            if kind in "012346789" or kind == "C2L" : # main ALU
                alu = codegen.Primitive("ALU", name)
                alu.params["ALU_MODE"] = kind
                if kind != "C2L":
                    alu.portmap['SUM'] = f"R{row}C{col}_F{idx}"
                alu.portmap['CIN'] = f"R{row}C{col}_CIN{idx}"
                alu.portmap['I2'] = f"R{row}C{col}_C{idx}"
                if idx != 5:
                    alu.portmap['COUT'] = f"R{row}C{col}_CIN{idx+1}"
                else:
                    alu.portmap['COUT'] = f"R{row}C{col + 1}_CIN{0}"
                if kind in "2346789":
                    alu.portmap['I0'] = f"R{row}C{col}_A{idx}"
                    alu.portmap['I1'] = f"R{row}C{col}_B{idx}"
                    if kind in "28":
                        alu.portmap['I3'] = f"R{row}C{col}_D{idx}"
                elif kind == "0":
                    alu.portmap['I0'] = f"R{row}C{col}_B{idx}"
                    alu.portmap['I1'] = f"R{row}C{col}_D{idx}"
                    alu.portmap['I3'] = f"R{row}C{col}_A{idx}"
                elif kind == "C2L":
                    alu.portmap['I0'] = f"R{row}C{col}_B{idx}"
                    alu.portmap['I1'] = f"R{row}C{col}_D{idx}"
                    alu.portmap['COUT'] = f"R{row}C{col}_F{idx}"
                    alu.params["ALU_MODE"] = "9" # XXX
                else:
                    alu.portmap['I0'] = f"R{row}C{col}_A{idx}"
                    alu.portmap['I1'] = f"R{row}C{col}_D{idx}"
                mod.wires.update(alu.portmap.values())
                mod.primitives[name] = alu
        elif typ == "RAM16":
            val0 = sum(1<<x for x in range(0,16) if not x in flags)
            val1 = sum(1<<(x-16) for x in range(16,32) if not x in flags)
            val2 = sum(1<<(x-32) for x in range(32,48) if not x in flags)
            val3 = sum(1<<(x-48) for x in range(48,64) if not x in flags)
            name = f"R{row}C{col}_RAM16"
            ram16 = codegen.Primitive("RAM16SDP4", name)
            ram16.params["INIT_0"] = f"16'b{val0:016b}"
            ram16.params["INIT_1"] = f"16'b{val1:016b}"
            ram16.params["INIT_2"] = f"16'b{val2:016b}"
            ram16.params["INIT_3"] = f"16'b{val3:016b}"
            ram16.portmap['DI'] = [f"R{row}C{col}_{x}5" for x in "DCBA"]
            ram16.portmap['CLK'] = f"R{row}C{col}_CLK2"
            ram16.portmap['WRE'] = f"R{row}C{col}_LSR2"
            ram16.portmap['WAD'] = [f"R{row}C{col}_{x}4" for x in "DCBA"]
            ram16.portmap['RAD'] = [f"R{row}C{col}_{x}0" for x in "DCBA"]
            ram16.portmap['DO'] = [f"R{row}C{col}_F{x}" for x in range(4, -1, -1)]
            mod.wires.update(chain.from_iterable([x if isinstance(x, list) else [x] for x in ram16.portmap.values()]))
            mod.primitives[name] = ram16
        elif typ in {"OSC", "OSCZ", "OSCF", "OSCH", "OSCW", "OSCO"}:
            name = f"R{row}C{col}_{typ}"
            osc = codegen.Primitive(typ, name)
            for paramval in flags:
                param, _, val = paramval.partition('=')
                osc.params[param] = val
            portmap = db[dbrow, dbcol].bels[bel].portmap
            for port, wname in portmap.items():
                osc.portmap[port] = f"R{row}C{col}_{wname}"
            mod.wires.update(osc.portmap.values())
            mod.primitives[name] = osc
        elif typ == "DFF":
            #print(flags)
            sd = False
            if 'SD' in flags:
                sd = True
                flags.remove('SD')
            kind, = flags # DFF only have one flag
            if kind == "RAM": continue
            idx = int(idx)
            if kind in latchmap:
                # Latch: gate signal uses CLK physical pin, adds E suffix for clock enable
                port = latchmap[kind]
                name = f"R{row}C{col}_DLE_{idx}"
                latch = codegen.Primitive(kind+"E", name)
                latch.portmap['CLK'] = f"R{row}C{col}_CLK{idx//2}"
                if sd:
                    latch.portmap['D'] = f"R{row}C{col}_SEL{idx}"
                else:
                    latch.portmap['D'] = f"R{row}C{col}_F{idx}"
                latch.portmap['Q'] = f"R{row}C{col}_Q{idx}"
                latch.portmap['CE'] = f"R{row}C{col}_CE{idx//2}"
                if port:
                    latch.portmap[port] = f"R{row}C{col}_LSR{idx//2}"
                mod.wires.update(latch.portmap.values())
                mod.primitives[name] = latch
                cst.cells[name] = (row, col, int(idx) // 2, _sides[int(idx) % 2])
            else:
                port = dffmap[kind]
                name = f"R{row}C{col}_{typ}E_{idx}"
                dff = codegen.Primitive(kind+"E", name)
                dff.portmap['CLK'] = f"R{row}C{col}_CLK{idx//2}"
                if sd:
                    dff.portmap['D'] = f"R{row}C{col}_SEL{idx}"
                else:
                    dff.portmap['D'] = f"R{row}C{col}_F{idx}"
                dff.portmap['Q'] = f"R{row}C{col}_Q{idx}"
                dff.portmap['CE'] = f"R{row}C{col}_CE{idx//2}"
                if port:
                    dff.portmap[port] = f"R{row}C{col}_LSR{idx//2}"
                mod.wires.update(dff.portmap.values())
                mod.primitives[name] = dff
                cst.cells[name] = (row, col, int(idx) // 2, _sides[int(idx) % 2])
        elif typ == "IOB":
            try:
                kind, = flags.intersection(iobmap.keys())
            except ValueError:
                continue
            flags.remove(kind)
            if kind == 'MIPI_IBUF':
                portmap = {}
                portmap['I'] = db[dbrow, dbcol].bels["IOBA"].portmap['I']
                portmap['IB'] = db[dbrow, dbcol].bels["IOBB"].portmap['I']
                portmap['OEN'] = db[dbrow, dbcol].bels["IOBA"].portmap['OE']
                portmap['OENB'] = db[dbrow, dbcol].bels["IOBB"].portmap['OE']
                portmap['OH'] = db[dbrow, dbcol].bels["IOBA"].portmap['O']
                portmap['OB'] = db[dbrow, dbcol].bels["IOBB"].portmap['O']
                portmap['HSREN'] = db[dbrow, dbcol].bels["IOLOGICB"].portmap['SETN']
                portmap['OL'] = db[dbrow, dbcol + 1].bels["IOBA"].portmap['O']
            elif kind == 'MIPI_OBUF':
                portmap = {}
                portmap['I'] = db[dbrow, dbcol].bels["IOBA"].portmap['I']
                portmap['IB'] = db[dbrow, dbcol].bels["IOBB"].portmap['I']
                portmap['IL'] = db[dbrow, dbcol].bels["IOBA"].portmap['I']
                portmap['MODESEL'] = db[dbrow, dbcol].bels["IOBA"].portmap['OE']
            elif kind == 'I3C_IOBUF':
                portmap = db[dbrow, dbcol].bels[bel].portmap.copy()
                portmap['MODESEL'] = portmap['OE']
                portmap.pop('OE', None)
            else:
                portmap = db[dbrow, dbcol].bels[bel].portmap
            name = f"R{row}C{col}_{kind}_{idx}"
            wires = set(iobmap[kind]['wires'])
            ports = set(chain.from_iterable(iobmap[kind].values())) - wires

            iob = codegen.Primitive(kind, name)

            if idx in iologic_detected:
                wires_suffix = '_IOL'
            else:
                wires_suffix = ''
            for port in wires:
                wname = portmap[port]
                if kind == 'MIPI_IBUF':
                    if port == 'OH':
                        wires_suffix_mipi = wires_suffix
                    else:
                        wires_suffix_mipi = ''
                    iob.portmap[portname(port)] = f"R{row}C{col}_{wname}{wires_suffix_mipi}"
                elif kind == 'MIPI_OBUF':
                    if port == 'I':
                        wires_suffix_mipi = wires_suffix
                    else:
                        wires_suffix_mipi = ''
                    iob.portmap[portname(port)] = f"R{row}C{col}_{wname}{wires_suffix_mipi}"
                else:
                    iob.portmap[portname(port)] = f"R{row}C{col}_{wname}{wires_suffix}"

            for port in ports:
                iob.portmap[port] = f"R{row}C{col}_{port}{idx}"

            wnames = [f"R{row}C{col}_{portmap[w]}" for w in iobmap[kind]['wires']]
            mod.wires.update(wnames)
            for direction in ['inputs', 'outputs', 'inouts']:
                wnames = [f"R{row}C{col}_{w}{idx}" for w in iobmap[kind].get(direction, [])]
                getattr(mod, direction).update(wnames)
            mod.primitives[name] = iob
            # constraints
            pos = chipdb.loc2pin_name(db, dbrow, dbcol)
            # XXX tangnano4k uses IOT30 not found in the package
            #bank = chipdb.loc2bank(db, dbrow, dbcol)
            cst.ports[name] = f"{pos}{idx}"
            if kind[0:5] == 'TLVDS':
                cst.ports[name] = f"{pos}{idx},{pos}{chr(ord(idx) + 1)}"
            #iostd = _banks.get(bank)
            #if iostd:
            #    cst.attrs.setdefault(name, {}).update({"IO_TYPE" : iostd})
            for flg in flags:
                name_val = flg.split('=')
                cst.attrs.setdefault(name, {}).update({name_val[0] : name_val[1]})

    # gnd = codegen.Primitive("GND", "mygnd")
    # gnd.portmap["G"] = "VSS"
    # mod.primitives["mygnd"] = gnd
    # vcc = codegen.Primitive("VCC", "myvcc")
    # vcc.portmap["V"] = "VCC"
    # mod.primitives["myvcc"] = vcc
    mod.assigns.append(("VCC", "1'b1"))
    mod.assigns.append(("VSS", "1'b0"))

def default_device_config():
    return {
        "JTAG regular_io":          "false",
        "SSPI regular_io":          "false",
        "MSPI regular_io":          "false",
        "READY regular_io":         "false",
        "DONE regular_io":          "false",
        "RECONFIG_N regular_io":    "false",
        "MODE regular_io":          "false",
        "CRC_check": "true",
        "compress": "false",
        "encryption": "false",
        "security_bit_enable": "true",
        "bsram_init_fuse_print": "true",
        "download_speed": "250/100",
        "spi_flash_address": "0x00FFF000",
        "format": "txt",
        "background_programming": "false",
        "secure_mode": "false"}

def fix_pll_ports(pll):
    for portname, up_limit in [('PSDA', 4), ('DUTYDA', 4), ('FDLY', 4), ('FBDSEL', 6), ('IDSEL', 6), ('ODSEL', 6)]:
        for n in range(0, up_limit):
            if f'{portname}{n}' in pll.portmap.keys():
                port = pll.portmap.setdefault(portname, [])
                port.append(pll.portmap[f'{portname}{n}'])
                pll.portmap.pop(f'{portname}{n}')

def fix_plls(db, mod):
    for pll_name, pll in [pr for pr in mod.primitives.items() if pr[1].typ in {'rPLL', 'PLLVR'}]:
        if 'INSEL' not in pll.params.keys():
            del mod.primitives[pll_name]
            continue
        disable_unused_pll_ports(pll)
        modify_pll_inputs(db, pll)
        mod.wires.update(pll.portmap.values())
        fix_pll_ports(pll)

def main():
    pil_available = True
    try:
        from PIL import Image
    except ImportError:
        pil_available = False

    parser = argparse.ArgumentParser(description='Unpack Gowin bitstream')
    parser.add_argument('bitstream')
    parser.add_argument('-d', '--device', required=True)
    parser.add_argument('-o', '--output', default='unpack.v')
    parser.add_argument('-s', '--cst', default=None)
    parser.add_argument('--noalu', action = 'store_true')
    if pil_available:
        parser.add_argument('--png')

    args = parser.parse_args()

    global _device
    _device = args.device
    # For tool integration it is allowed to pass a full part number
    m = re.match("GW1N(S?)[A-Z]*-(LV|UV|UX)([0-9])C?([A-Z]{2}[0-9]+P?)(C[0-9]/I[0-9])", _device)
    if m:
        mods = m.group(1)
        luts = m.group(3)
        _device = f"GW1N{mods}-{luts}"

    with importlib.resources.path('apycula', f'{args.device}.msgpack.xz') as path:
        db = load_chipdb(path)

    global _pinout
    _pinout = db.pinout[_device][_packages[_device]]

    bitmap, _, _, extra_slots = read_bitstream(args.bitstream)
    bm = chipdb.tile_bitmap(db, bitmap)
    mod = codegen.Module()
    cst = codegen.Constraints()

    if pil_available and args.png:
        display(args.png, bitmap)

    # make wire aliases from Himbaechel nodes
    def by_name_len(el):
        return len(el[2])

    for node_desc in db.nodes.values():
        root_wire = None
        for row, col, wire in sorted(node_desc[1], key = by_name_len):
            wire_name = f'R{row + 1}C{col + 1}_{wire}'
            if not root_wire:
                root_wire = wire_name
                continue
            mod.wire_aliases[wire_name] = root_wire
    for row in range(db.rows):
        for col in range(db.cols):
            for i in [1, 2]:
                mod.wire_aliases[chipdb.wire2global(row + 0, col + 1, db, f'N1{i}1')] = f'R{row + 1}C{col + 1}_SN{i}0'
                mod.wire_aliases[chipdb.wire2global(row + 2, col + 1, db, f'S1{i}1')] = f'R{row + 1}C{col + 1}_SN{i}0'
                mod.wire_aliases[chipdb.wire2global(row + 1, col + 0, db, f'W1{i}1')] = f'R{row + 1}C{col + 1}_EW{i}0'
                mod.wire_aliases[chipdb.wire2global(row + 1, col + 2, db, f'E1{i}1')] = f'R{row + 1}C{col + 1}_EW{i}0'

    # Slots have no wires only func fuses
    if extra_slots:
        for slot_idx, slot_bitmap in extra_slots.items():
            if slot_idx in {2, 3, 4, 5, 6, 8}:
                av = parse_attrvals(slot_bitmap, db.rev_logicinfo('PLL'), db.shortval[1024]['PLL'], attrids.pll_attrids, "PLL")
                print('Slot:', slot_idx, av)
            elif slot_idx in {1}:
                av = parse_attrvals(slot_bitmap, db.rev_logicinfo('ADC'), db.shortval[1026]['ADC'], attrids.adc_attrids, "ADC")
                print('Slot:', slot_idx, av)
            else:
                print('Unknown Slot:', slot_idx)

    # XXX this PLLs have empty main cell
    if _device in {'GW1N-9C', 'GW1N-9'}:
        bm_pll = chipdb.tile_bitmap(db, bitmap, empty = True)
        bm[(9, 0)] = bm_pll[(9, 0)]
        bm[(9, 46)] = bm_pll[(9, 46)]
    if _device in {'GW2A-18', 'GW2A-18C'}:
        bm_pll = chipdb.tile_bitmap(db, bitmap, empty = True)
        bm[(9, 0)] = bm_pll[(9, 0)]
        bm[(9, 55)] = bm_pll[(9, 55)]
        bm[(45, 0)] = bm_pll[(45, 0)]
        bm[(45, 55)] = bm_pll[(45, 55)]

    # banks first: need to know iostandards
    for pos in db.bank_tiles.values():
        row, col = pos
        try:
            t = bm[(row, col)]
        except KeyError:
            continue
        bels, pips, clock_pips = parse_tile_(db, row, col, t, bm)
        #print("bels:", bels)
        tile2verilog(row, col, bels, pips, clock_pips, mod, cst, db)

    for idx, t in bm.items():
        row, col = idx
        # skip banks
        if (row, col) in db.bank_tiles.values():
            continue
        bels, pips, clock_pips = parse_tile_(db, row, col, t, bm, noiostd = False)
        #print("bels:", idx, bels)
        #print(pips)
        #print(clock_pips)
        if args.noalu:
            removeALUs(bels)
        else:
            removeLUTs(bels)
        ram16_remove_bels(bels)
        tile2verilog(row, col, bels, pips, clock_pips, mod, cst, db)

    fix_plls(db, mod)

    with open(args.output, 'w') as f:
        mod.write(f)

    if args.cst:
        with open(args.cst, 'w') as f:
            cst.write(f)

if __name__ == "__main__":
    main()



from os.path import expanduser
from glob import glob
import json
import os
import csv

VeryTrue = 2

# caches
# .CSV index of vendor files {(device, package) : file_name}
_pindef_index = {}
# (device, package) : pins
_pindef_files = {}

def get_package(device, package, special_pins):
    global _pindef_files
    if (device, package) not in _pindef_files:
        gowinhome = os.getenv("GOWINHOME")
        if not gowinhome:
            raise Exception("GOWINHOME not set")
        with open(_pindef_index[(device, package)]) as f:
            pins = json.load(f)
        _pindef_files[(device, package)] = [d for d in pins['PIN_DATA'] if d['TYPE'] == 'I/O']

    if special_pins != VeryTrue:
        pins = [pin for pin in _pindef_files[(device, package)]
                if 'CFG' not in pin.keys() or (
                    pin['CFG'] != 'RECONFIG_N' and not pin['CFG'].startswith('JTAGSEL_N'))]
    else:
        pins = _pindef_files[(device, package)]
    if not special_pins:
        return [pin for pin in pins if 'CFG' not in pin.keys()]
    return pins

# {partnumber : (pkg, device, speed)}
def all_packages(device):
    gowinhome = os.getenv("GOWINHOME")
    if not gowinhome:
        raise Exception("GOWINHOME not set")
    # {package: speed} vendor file
    speeds = {}
    with open(f"{gowinhome}/IDE/data/device/device_info.csv", mode='r') as csv_file:
        csv_reader = csv.DictReader(csv_file, fieldnames =
            ["unused_id", "partnumber", "series", "device", "unused_0", "unused_1", "package", "voltage", "speed"])
        for row in csv_reader:
            if row['device'] != device:
               continue
            speeds.update({row['partnumber']: row['speed']})
    global _pindef_index
    # _pindef_index = {}
    res = {}
    with open(f"{gowinhome}/IDE/data/device/device_package.csv", mode='r') as csv_file:
        csv_reader = csv.DictReader(csv_file, fieldnames =
            ["unused_id", "partnumber", "series", "device", "package", "filename"])
        for row in csv_reader:
            if row['device'] != device:
               continue
            res[row['partnumber']] = (row['package'], device, speeds[row['partnumber']])
            _pindef_index[(row['device'], row['package'])] = \
                    f"{gowinhome}/IDE/data/device/{row['filename']}"
    return res

def get_pins(device, package, special_pins=False):
    df = get_package(device, package, special_pins)
    res = {}
    for pin in df:
        res.setdefault(str(pin['BANK']), []).append(str(pin['INDEX']))
    return res

def get_bank_pins(device, package):
    df = get_package(device, package, VeryTrue)
    res = {}
    for pin in df:
        res[pin['NAME']] = str(pin['BANK'])
    return res

def get_locs(device, package, special_pins=False):
    df = get_package(device, package, special_pins)
    res = set()
    for pin in df:
        res.update({pin['NAME']})
    return res

def get_pin_locs(device, package, special_pins=False):
    df = get_package(device, package, special_pins)
    res = {}
    for pin in df:
        cfgs = []
        if 'CFG' in pin.keys():
            cfgs = pin['CFG'].split('/')
        res[str(pin['INDEX'])] = (pin['NAME'], cfgs)
    return res

def get_clock_locs(device, package):
    df = get_package(device, package, True)
    return [(pin['NAME'], *pin['CFG'].split('/')) for pin in df
            if 'CFG' in pin.keys() and pin['CFG'].startswith("GCLK")]

def get_pll_pads_locs(device, package):
    df = get_package(device, package, True)
    return [(pin['NAME'], *pin['CFG'].split('/')) for pin in df
            if 'CFG' in pin.keys() and 'PLL' in pin['CFG']]

# { name : (is_diff, is_true_lvds, is_positive, adc_bus)}
def get_diff_adc_cap_info(device, package, special_pins=False):
    df = get_package(device, package, special_pins)
    res = {}
    # If one pin of the pair is forbidden for the diff IO,
    # we can determine this only after we read the data of all pairs
    positive = {}
    negative = {}
    for pin in df:
        is_positive = False
        is_diff = 'DIFF' in pin.keys()
        adc_bus = None
        if 'ADC_INPUT' in pin.keys() and pin['ADC_INPUT']:
            adc_bus = pin['ADC_INPUT']

        if not is_diff:
            res[str(pin['NAME'])] = (is_diff, is_true_lvds, is_positive, adc_bus)
            continue
        is_true_lvds = 'TRUELVDS' in pin.keys()

        if pin['DIFF'] == 'P':
            is_positive = True
            positive[str(pin['NAME'])] = (is_diff, is_true_lvds, is_positive, adc_bus, str(pin['PAIR']))
        else:
            is_positive = False
            negative[str(pin['NAME'])] = (is_diff, is_true_lvds, is_positive, adc_bus)
    # check the pairs
    for pos_name, pos_flags in positive.items():
        neg_name = pos_flags[-1]
        if neg_name in negative.keys():
            res.update({pos_name : pos_flags[0:-1]})
            res.update({neg_name : negative[neg_name]})
    return res





try:
    import numpy as np

    def fliplr(bmp):
        return np.fliplr(np.asarray(bmp))

    def flipud(bmp):
        return np.flipud(np.asarray(bmp))

    def transpose(bmp):
        return np.asarray(bmp).T

    def vstack(bmp_0, bmp_1):
        return np.vstack((np.asarray(bmp_0), np.asarray(bmp_1)))

    def hstack(bmp_0, bmp_1):
        return np.hstack((np.asarray(bmp_0), np.asarray(bmp_1)))

    def shape(bmp):
        return np.asarray(bmp).shape

    def fill(rows, cols, val):
        return np.full((rows, cols), val, dtype=np.uint8)

    def ones(rows, cols):
        return np.ones((rows, cols), dtype=np.uint8)

    def zeros(rows, cols):
        return np.zeros((rows, cols), dtype=np.uint8)

    def packbits(bmp, axis=None):
        return np.packbits(np.asarray(bmp, dtype=np.uint8), axis=axis)

    def bsum(arr):
        return np.asarray(arr).sum()

    def xor(bmp_0, bmp_1):
        return np.bitwise_xor(np.asarray(bmp_0), np.asarray(bmp_1))

    def any(bmp):
        return np.any(np.asarray(bmp))

    def nonzero(bmp):
        return np.nonzero(np.asarray(bmp))

    def byte_histogram(lst):
        return np.bincount(np.asarray(lst).ravel(), minlength=256)

    def histogram(lst, bins):
        return np.histogram(np.asarray(lst), bins=bins)[0]

    def blit(dst, y, x, src):
        src = np.asarray(src)
        h, w = src.shape
        dst[y:y+h, x:x+w] = src

except ImportError:
    import warnings
    warnings.warn("Numpy is not available, performance will be degraded.")
    
    def fliplr(bmp):
        """
        Flips the entries in each row in the left/right direction.
        Returns a new matrix.
        """
        return [row[::-1] for row in bmp]

    def flipud(bmp):
        """
        Reverse the order of elements in each column (up/down).
        Returns a refence.
        """
        return bmp[::-1]

    def transpose(bmp):
        """
        Transposes a bitmap (swaps rows and cols)
        Returns a reference
        """
        return [[bmp[j][i] for j in range(len(bmp))] for i in range(len(bmp[0]))]

    def vstack(bmp_0, bmp_1):
        """
        Stack matrices in sequence vertically (row wise).
        Returns a reference.
        """
        return [*bmp_0, *bmp_1]

    def hstack(bmp_0, bmp_1):
        """
        Stack matrices in sequence horizontally (column wise).
        Returns a new matrix.
        """
        return [bmp[0] + bmp[1] for bmp in zip(bmp_0, bmp_1)]

    def shape(bmp):
        """
        Return the shape of a matrix.
        """
        return [len(bmp), len(bmp[0])]

    def fill(rows, cols, val):
        """
        Returns a new matrix of given shape, filled with val.
        """
        return [[val] * cols for i in range(rows)]

    def ones(rows, cols):
        """
        Returns a new matrix of given shape, filled with ones.
        """
        return fill(rows, cols, 1)

    def zeros(rows, cols):
        """
        Returns a new matrix of given shape, filled with zeros.
        """
        return fill(rows, cols, 0)

    def packbits(bmp, axis=None):
        """
        Packs the elements of a bitmap into bytes.
        [1, 1, 0, 0, 0] -> [24]  # [5'b11000]
        Returns a list of bytes.
        """
        byte_list = []
        byte = 0
        bit_cnt = 0
        if not axis:
            for bmp_r in bmp:
                for col in range(shape(bmp)[1]):
                    byte = (byte << 1) + bmp_r[col]
                    bit_cnt += 1
                    if bit_cnt == 8:
                        byte_list.append(byte)
                        bit_cnt = 0
                        byte = 0
        else:
            for bmp_r in bmp:
                byte_list.append([])
                byte_list_r = byte_list[-1]
                for col in range(shape(bmp)[1]):
                    byte = (byte << 1) + bmp_r[col]
                    bit_cnt += 1
                    if bit_cnt == 8:
                        byte_list_r.append(byte)
                        bit_cnt = 0
                        byte = 0
        return byte_list

    def bsum(arr):
        return sum(arr)

    def xor(bmp_0, bmp_1):
        """
        Bitwise XOR
        Returns a new matrix
        """
        return [[ vals[0] ^ vals[1]for vals in zip(row[0], row[1])] for row in zip(bmp_0, bmp_1)]

    def histogram(lst, bins):
        """
        Compute the histogram of a list.
        Returns a list of counters.
        """
        l_bins = len(bins) - 1
        r_lst = [0] * l_bins
        from itertools import chain
        for val in chain.from_iterable(lst):
            for i in range(l_bins):
                if val in range(bins[i], bins[i + 1]) or (i == l_bins - 1 and val == bins[-1]):
                    r_lst[i] += 1
        return r_lst

    def byte_histogram(lst):
        """
        Compute the histogram of a list of bytes.
        Returns a list of 256 counters.
        """
        r_lst = [0] * 256
        from itertools import chain
        for val in chain.from_iterable(lst):
            r_lst[val] += 1
        return r_lst

    def any(bmp):
        """
        Test whether any matrix element evaluates to True.
        """
        for row in bmp:
            for val in row:
                if val:
                    return True
        return False

    def nonzero(bmp):
        """
        Return the indices of the elements that are non-zero.
        """
        res = ([], [])
        for ri, row in enumerate(bmp):
            for ci, val in enumerate(row):
                if val:
                    res[0].append(ri)
                    res[1].append(ci)
        return res

    def blit(dst, y, x, src):
        for row in src:
            x0 = x
            for val in row:
                dst[y][x0] = val
                x0 += 1
            y += 1



from write_fasm import *
import re
from apycula import codegen

# Need to tell FASM generator how to write parameters
# (celltype, parameter) -> ParameterConfig
param_map = {
	("GENERIC_SLICE", "K"): ParameterConfig(write=False),
	("GENERIC_SLICE", "INIT"): ParameterConfig(write=True, numeric=True, width=2**4),
	("GENERIC_SLICE", "FF_USED"): ParameterConfig(write=True, numeric=True, width=1),

	("GENERIC_IOB", "INPUT_USED"): ParameterConfig(write=True, numeric=True, width=1),
	("GENERIC_IOB", "OUTPUT_USED"): ParameterConfig(write=True, numeric=True, width=1),
	("GENERIC_IOB", "ENABLE_USED"): ParameterConfig(write=True, numeric=True, width=1),
}

#gen_7_ PLACE_IOT8[B] //IBUF
#gen_4_ PLACE_R13C6[0][A]
def write_posp(f):
    belre = re.compile(r"R(\d+)C(\d+)_(SLICE|IOB)(\w)")
    namere = re.compile(r"\W+")
    for name, cell in ctx.cells:
        row, col, typ, idx = belre.match(cell.bel).groups()
        row = int(row)
        col = int(col)
        name = namere.sub('_', name)
        if typ == 'SLICE':
            idx = int(idx)
            cls = idx//2
            side = ['A', 'B'][idx%2]
            lutname = name + "_LUT"
            f.write(f"{lutname} PLACE_R{row}C{col}[{cls}][{side}]\n")

            lut = codegen.Primitive("LUT4", lutname)
            #lut.params["INIT"] = f"16'b{val:016b}"
            lut.portmap['F'] = f"R{row}C{col}_F{idx}"
            lut.portmap['I0'] = f"R{row}C{col}_A{idx}"
            lut.portmap['I1'] = f"R{row}C{col}_B{idx}"
            lut.portmap['I2'] = f"R{row}C{col}_C{idx}"
            lut.portmap['I3'] = f"R{row}C{col}_D{idx}"
            mod.wires.update(lut.portmap.values())
            mod.primitives[lutname] = lut

            if int(cell.params['FF_USED'], 2):
                dffname = name + "_DFF"
                f.write(f"{dffname} PLACE_R{row}C{col}[{cls}][{side}]\n")

                lut = codegen.Primitive("DFF", dffname)
                lut.portmap['D'] = f"R{row}C{col}_F{idx}"
                lut.portmap['Q'] = f"R{row}C{col}_Q{idx}"
                lut.portmap['CLK'] = f"R{row}C{col}_CLK{idx}"
                mod.wires.update(lut.portmap.values())
                mod.primitives[dffname] = lut
        elif typ == 'IOB':
            if row == 1:
                edge = 'T'
                num = col
            elif col == 1:
                edge = 'L'
                num = row
            elif col == 47: #TODO parameterize
                edge = 'R'
                num = row
            else:
                edge = 'B'
                num = col
            f.write(f"{name} PLACE_IO{edge}{num}[{idx}]\n")

            iob = codegen.Primitive("IOBUF", name)
            iob.portmap['I'] = f"R{row}C{col}_I{idx}"
            iob.portmap['O'] = f"R{row}C{col}_O{idx}"
            iob.portmap['IO'] = f"R{row}C{col}_IO{idx}"
            iob.portmap['OEN'] = f"R{row}C{col}_OEN{idx}"
            mod.wires.update(iob.portmap.values())
            mod.inouts.add(f"R{row}C{col}_IO{idx}")
            mod.primitives[name] = iob


with open("blinky.fasm", "w") as f:
	write_fasm(ctx, param_map, f)

mod = codegen.Module()
with open("blinky.posp", "w") as f:
    write_posp(f)

with open("blinky.vm", "w") as f:
    mod.write(f)

#code.interact(local=locals())



import re, csv, os, sys
from enum import Enum
from pathlib import Path
import argparse

"""
Known IO Features
    name -> string representation
    count -> -1 if this feature has only a single option, the size (in bytes) of the encoding of number of options otherwise
    size -> number of bytes used to encode an option
    list -> True if this feature has multiple configuration options, false otherwise
"""
class IOF(Enum):
    TYPE = {"name":"Type", "count":-1, "size":4, "list":False}
    DRIVE = {"name": "Drive", "count": 2, "size":4, "list":True}
    DIFF_DRIVE = {"name": "Differential Drive", "count": 2, "size":4, "list":True}
    OPEN_DRAIN = {"name": "Open Drain", "count": 2, "size":4, "list":True}
    SLEW_RATE = {"name": "Slew Rate", "count": 2, "size":4, "list":True}
    CLAMP = {"name": "Clamp", "count": 2, "size":4, "list":True}
    VREF = {"name": "Voltage Reference", "count": 2, "size":4, "list":True}
    HYSTERISIS = {"name": "Hysterisis", "count": 2, "size":4, "list":True}
    PULL_MODE = {"name": "Pull Mode", "count": 2, "size":4, "list":True}
    SINGLE_RES = {"name": "Single Resistor", "count": 2, "size":4, "list":True}
    DIFF_RES = {"name": "Differential Resistor", "count": 2, "size":4, "list":True}
    VCC_IO = {"name": "VCC IO", "count": 2, "size":4, "list":True}
    I3C_MODE = {"name": "I3C Mode", "count": 2, "size":4, "list":True}
    MIPI_INPUT = {"name": "MIPI Input", "count": 2, "size":4, "list":True}
    MIPI_OUTPUT = {"name": "MIPI Output", "count": 2, "size":4, "list":True}
    PULL_STRENGTH = {"name": "PULL STRENGTH", "count": 2, "size":4, "list":True}
    BANK = {"name": "Bank", "count": 2, "size":4, "list":True}
    MODE = {"name": "Mode", "count": -1, "size":4, "list":False}


"""
Structure of Known .ini files
    align_start -> number of bytes used to encode the start alignment
    features -> *ordered* list of configurable IO features
    input -> number of input IO configuration points (each configuration point has options for all features)
    output -> number of output IO configuration points
    bidirectional -> number of bidirectional IO configuration points
    i3c_bank -> True if the device has an i3c_bank, False otherwise
    align_end -> number of bytes used to encode the end alignment
"""
schemas = [
    {
        "devices": ['GW1N-1', 'GW1NR-1'],
        "align_start": 2, 
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.PULL_MODE, IOF.SINGLE_RES, IOF.DIFF_RES, IOF.VCC_IO, IOF.MODE],
        "input": 0x3b,
        "output": 0x25,
        "bidirectional": 0x25,
        "i3c_bank": False,
        "align_end": 2
    },

     {
        "devices": ['GW1N-4', 'GW1NR-4', 'GW1NRF-4', 'GW1N-4D', 'GW1NR-4D'],
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.PULL_MODE, IOF.SINGLE_RES, IOF.DIFF_RES, IOF.VCC_IO, IOF.MODE],
        "input": 0x3b,
        "output": 0x29,
        "bidirectional": 0x29,
        "i3c_bank": False,
        "align_end": 2
    },

    {
        "devices": ['GW1N-2', 'GW1NZ-2', 'GW1NR-2', 'GW1NZR-2', 'GW1N-1P5', 'GW1N-2B', 'GW1NR-2B', 'GW1N-1P5B', 'GW1N-2C', 'GW1NR-2C', 'GW1N-1P5C', 'GW1NZ-2B', 'GW1NZ-2C'], 
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.I3C_MODE, IOF.MIPI_INPUT, IOF.MIPI_OUTPUT, IOF.PULL_MODE, IOF.SINGLE_RES,
                    IOF.DIFF_RES, IOF.VCC_IO, IOF.MODE],
        "input": 0x3c,
        "output": 0x2a,
        "bidirectional": 0x26,
        "i3c_bank": True,
        "align_end": 2
    },
    
    {
        "devices": ['GW1N-6', 'GW1N-9', 'GW1NR-9', 'GW1N-9C', 'GW1NR-9C', 'GW1NS-2', 'GW1NSR-2'],
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.I3C_MODE, IOF.MIPI_INPUT, IOF.MIPI_OUTPUT, IOF.PULL_MODE, IOF.SINGLE_RES,
                    IOF.DIFF_RES, IOF.VCC_IO, IOF.MODE],
        "input": 0x3c,
        "output": 0x2a,
        "bidirectional": 0x26,
        "i3c_bank": True,
        "align_end": 2
    },

    {
        "devices": ['GW1NZ-1', 'GW1NZ-1C'],
        "features": [IOF.TYPE, IOF.DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.HYSTERISIS, IOF.PULL_MODE, 
                     IOF.VCC_IO, IOF.MODE],
        "input": 0x18,
        "output": 0x0c,
        "bidirectional": 7,
        "i3c_bank": False,
        "align_end": 2
    },

    {
        "devices": ['GW1N-1S'], 
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.MIPI_INPUT, IOF.PULL_MODE, IOF.SINGLE_RES, IOF.DIFF_RES, IOF.VCC_IO],
        "input": 0x38,
        "output": 0x25,
        "bidirectional": 0x25,
        "i3c_bank": False,
        "align_end": 2

    },

     {
        "devices": ['GW1NS-4', 'GW1NSR-4'], 
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.I3C_MODE, IOF.MIPI_INPUT, IOF.MIPI_OUTPUT, IOF.PULL_MODE, IOF.SINGLE_RES, 
                     IOF.DIFF_RES, IOF.VCC_IO, IOF.MODE],
        "input": 0x3c,
        "output": 0x2a,
        "bidirectional": 0x26,
        "i3c_bank": True,
        "align_end": 2
    }, 

    {
        "devices": ['GW2A-18', 'GW2A-55', 'GW2AR-18', 'GW2A-55C', 'GW2A-18C', 'GW2AR-18C', 'GW2ANR-18C', 'GW2AN-55C'], 
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.PULL_MODE, IOF.SINGLE_RES, IOF.DIFF_RES, IOF.VCC_IO, IOF.MODE],
        "input": 0x3b, 
        "output": 0x29,
        "bidirectional": 0x29,
        "i3c_bank": False,
        "align_end": 2
    },

    {
        "devices": ['GW5AT-138'], 
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.I3C_MODE, IOF.MIPI_INPUT, IOF.MIPI_OUTPUT, IOF.PULL_MODE, IOF.SINGLE_RES, 
                     IOF.DIFF_RES, IOF.VCC_IO, IOF.MODE],
        "input": 0x39,
        "output": 0x2e,
        "bidirectional": 0x2c,
        "i3c_bank": True,
        "align_end": 2
    },

    {
        "devices": ['GW2AN-18X', 'GW2AN-9X'], 
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.I3C_MODE, IOF.MIPI_INPUT, IOF.PULL_MODE, IOF.SINGLE_RES, IOF.DIFF_RES, 
                     IOF.VCC_IO, IOF.MODE],
        "input": 0x3c,
        "output": 0x29,
        "bidirectional": 0x25,
        "i3c_bank": True,
        "align_end": 2
    },

    {
        "devices": ['GW5AT-138B', 'GW5AST-138B', 'GW5A-138B', 'GW5AS-138B'], 
        "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.I3C_MODE, IOF.MIPI_INPUT,  IOF.MIPI_OUTPUT, IOF.PULL_STRENGTH, IOF.PULL_MODE, 
                     IOF.SINGLE_RES, IOF.DIFF_RES, IOF.BANK,  IOF.VCC_IO, IOF.MODE],
        "input": 0x3b,
        "output": 0x2e,
        "bidirectional": 0x2c,
        "i3c_bank": True,
        "align_end": 2
    },

    {
        "devices": ['GW5AT-60'], 
         "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.I3C_MODE, IOF.MIPI_INPUT,  IOF.MIPI_OUTPUT, IOF.PULL_STRENGTH, IOF.PULL_MODE, 
                     IOF.SINGLE_RES, IOF.DIFF_RES, IOF.BANK,  IOF.VCC_IO, IOF.MODE],
        "input": 0x52,
        "output": 0x37,
        "bidirectional": 0x36,
        "i3c_bank": True,
        "align_end": 2
    },

    {
        "devices": ['GW5A-25A', 'GW5AR-25A'], 
         "features": [IOF.TYPE, IOF.DRIVE, IOF.DIFF_DRIVE, IOF.OPEN_DRAIN, IOF.SLEW_RATE, IOF.CLAMP, IOF.VREF, 
                     IOF.HYSTERISIS, IOF.I3C_MODE, IOF.MIPI_INPUT,  IOF.MIPI_OUTPUT, IOF.PULL_STRENGTH, IOF.PULL_MODE, 
                     IOF.SINGLE_RES, IOF.DIFF_RES, IOF.BANK,  IOF.VCC_IO, IOF.MODE],
        "input": 0x52,
        "output": 0x37,
        "bidirectional": 0x36,
        "i3c_bank": True,
        "align_end": 2
    }
]

"""
Known mapping of PIO binary-file options to strings
"""
PIO_STR_MAPPINGS = {0: 'UNKNOWN', 1: 'MIPI', 2: 'BLVDS25E', 3: 'BLVDS25', 4: 'BLVDS_E', 5: 'HSTL', 6: 'HSTL_D', 7: 'HSTL15_I', 8: 'HSTL15D_I', 
                   9: 'HSTL18_I', 10: 'HSTL18_II', 11: 'HSTL18D_I', 12: 'HSTL18D_II', 13: 'SSTL', 14: 'SSTL_D', 15: 'SSTL15', 16: 'SSTL15D', 
                   17: 'SSTL18_I', 18: 'SSTL18_II', 19: 'SSTL18D_I', 20: 'SSTL18D_II', 21: 'SSTL25_I', 22: 'SSTL25_II', 23: 'SSTL25D_I', 
                   24: 'SSTL25D_II', 25: 'SSTL33_I', 26: 'SSTL33_II', 27: 'SSTL33D_I', 28: 'SSTL33D_II', 29: 'LVCMOS12', 30: 'LVCMOS15', 
                   31: 'LVCMOS18', 32: 'LVCMOS25', 33: 'LVCMOS33', 34: 'LVCMOS_D', 35: 'LVCMOS12D', 36: 'LVCMOS15D', 37: 'LVCMOS18D', 
                   38: 'LVCMOS25D', 39: 'LVCMOS33D', 40: 'LVDS', 41: 'LVDS_E', 42: 'LVDS25', 43: 'LVDS25E', 44: 'LVPECL33', 45: 'LVPECL33E', 
                   46: 'LVTTL33', 47: 'MLVDS25', 48: 'MLVDS_E', 49: 'MLVDS25E', 50: 'RSDS25E', 51: 'PCI33', 52: 'RSDS', 53: 'RSDS25', 
                   54: 'RSDS_E', 55: 'MINILVDS', 56: 'PPLVDS', 57: 'VREF1_DRIVER', 58: 'VREF2_DRIVER', 59: 'LVCMOS33OD25', 60: 'LVCMOS33OD18', 
                   61: 'LVCMOS33OD15', 62: 'LVCMOS25OD18', 63: 'LVCMOS25OD15', 64: 'LVCMOS18OD15', 65: 'LVCMOS15OD12', 66: 'LVCMOS25UD33', 
                   67: 'LVCMOS18UD25', 68: 'LVCMOS18UD33', 69: 'LVCMOS15UD18', 70: 'LVCMOS15UD25', 71: 'LVCMOS15UD33', 72: 'LVCMOS12UD15', 
                   73: 'LVCMOS12UD18', 74: 'LVCMOS12UD25', 75: 'LVCMOS12UD33', 76: 'VREF1_LOAD', 77: 'VREF2_LOAD', 78: 'ENABLE', 79: 'TRIMUX', 
                   80: 'PADDI', 81: 'PGBUF', 82: '0', 83: '1', 84: 'SIG', 85: 'INV', 86: 'TO', 87: '1.2', 88: '1.25', 89: '1.5', 90: '1.8', 
                   91: '2.0', 92: '2.5', 93: '3.3', 94: '3.5', 95: '2', 96: '4', 97: '6', 98: '8', 99: '12', 100: '16', 101: '20', 102: '24', 
                   103: '80', 104: '100', 105: '120', 106: 'NA', 107: 'ON', 108: 'OFF', 109: 'PCI', 110: 'HIGH', 111: 'H2L', 112: 'L2H', 
                   113: 'DOWN', 114: 'KEEPER', 115: 'NONE', 116: 'UP', 117: 'FAST', 118: 'SLOW', 119: 'I45', 120: 'I50', 121: 'I55', 122: 'TSREG', 
                   123: 'TMDDR', 124: 'OD1', 125: 'OD2', 126: 'OD3', 127: 'UD1', 128: 'UD3', 129: 'INTERNAL', 130: 'SINGLE', 131: 'DIFF', 
                   132: 'IN12', 133: 'UD2', 134: 'LVPECL_E', 135: '68', 136: '3', 137: '5', 138: '7', 139: '9', 140: '10', 141: '11', 142: '4.5', 
                   143: 'MIPI_IBUF', 144: '1.35', 145: '5.5', 146: '6.5', 147: '10.5', 148: '13.5', 149: '14', 150: 'TMDS33', 151: 'LPDDR', 
                   152: 'HSUL12', 153: 'HSUL12D', 154: 'HSTL12_I', 155: 'HSTL15_II', 156: 'HSTL15D_II', 157: 'SSTL12', 158: 'SSTL135', 
                   159: 'SSTL135D', 160: 'LVCMOS10', 161: 'LVCMOS33OD12', 162: 'LVCMOS25OD12', 163: 'LVCMOS18OD12', 164: 'HSTL12D_I', 
                   165: 'LVCMOS10D', 166: 'LVCMOS10UD12', 167: 'LVCMOS10UD15', 168: 'LVCMOS10UD18', 169: 'LVCMOS10UD25', 170: 'LVCMOS10UD33', 
                   171: 'LVCMOS15OD10', 172: 'LVCMOS18OD10', 173: '40', 174: '50', 175: '60', 176: '75', 177: '0.333VCCIO', 178: '0.417VCCIO', 
                   179: '0.583VCCIO', 180: '0.6V', 181: '0.675V', 182: '0.75V', 183: '0.9V', 184: '1.25V', 185: '1.5V', 186: 'MEDIUM', 187: 'WEAK', 
                   188: 'STRONG', 189: 'VCCX', 190: '1.0', 191: '1.375', 192: '134', 193: '139', 194: '145', 195: '150', 196: '72', 197: '74', 
                   198: '77', 199: '85', 200: '86', 201: '89', 202: '91', 203: '106', 204: '108', 205: '112', 206: '115', 207: '128', 208: '132', 
                   209: '172', 210: '180', 211: '82', 212: '84', 213: '99', 214: '101', 215: '123', 216: '164', 217: '96', 218: '157', 219: '79', 
                   220: '93', 221: '124', 222: '70', 223: '88', 224: '92', 225: '94', 226: '98', 227: '78', 228: '76', 229: '118', 235: '102', 
                   236: 'LPDDRD', 238: 'LVCMOS25OD10', 239: 'LVCMOS33OD10', 240: 'LVCMOS12OD10', 241: 'SSTL2D_I', 242: 'SSTL2D_II', 243: 'SSTL3D_I', 
                   244: 'SSTL3D_II', 245: 'SSTL12D_I', 246: 'SSTL15D_I', 247: 'SSTL135D_I', 248: 'SSTL2_I', 249: 'SSTL2_II', 250: 'SSTL3_I', 
                   251: 'SSTL3_II', 252: 'SSTL12_I', 253: 'SSTL15_I', 254: 'SSTL135_I', 255: 'ADC_IN', 256: '25', 257: '90', 258: '122', 259: '130', 
                   260: '0.5VCCIO', 261: '0.643VCCIO', 262: '0.357VCCIO', 298: 'LOW', 299: 'UHIGH'}

class IniParser:
    def __init__ (self, device:str, schema:dict=None, ini_file:str|Path=None, pio_str_mappings:Path|dict[int: str]=PIO_STR_MAPPINGS):
        self.device = device

        if schema is None:
            schema_options = [x for x in schemas if device.upper() in x["devices"]]
            if schema_options:
                self.schema = schema_options[0]
            else:
                self.schema = None
        else:
            self.schema = schema

        if not self.schema:
            raise Exception(f"No schema supplied/found for device {self.device}")
        
        self.ini_file = ini_file
        self.input = {}
        self.output = {}
        self.bidirectional = {}
        self.i3c_bank = []
        self.__pointer = 0
        self.byte_array = []

        if isinstance(pio_str_mappings, str) or isinstance(pio_str_mappings, Path):

            self.pio_str_mappings = {}
            self.str_pio_mappings = {}
            with open (pio_str_mappings, "r") as f:
                reader = csv.reader(f)
                for line in reader:
                    id, string, id_hex = line
                    id = int(id)
                    self.pio_str_mappings[id] = string
                    self.str_pio_mappings[string] = id

        elif isinstance(pio_str_mappings, dict):
            self.pio_str_mappings = pio_str_mappings
            self.str_pio_mappings = {v: k for k, v in self.pio_str_mappings.items()}
    
    def parse(self, ini_file=None):
        self.__pointer=0

        if ini_file is None or not ini_file:
            gowinhome = os.getenv("GOWINHOME")
            if not gowinhome:
                raise Exception("GOWINHOME not set")
            ini_file = f"{gowinhome}/IDE/share/device/{self.device}/{self.device}.ini"
    
        with open(ini_file, 'rb') as f:
            self.byte_array = f.read() #The ini files are pretty small so this is fine
            features = self.schema["features"]
            input_lines = self.schema["input"]
            output_lines = self.schema["output"]
            bidirectional_lines = self.schema["bidirectional"]
            i3c_bank = self.schema["i3c_bank"]

            start_align = self.__read_val(word_size=2, byteorder="little")
            assert start_align==1, "Start alignment incorrect, parsing failed"
            
            for idx in range(input_lines):
                self.input[idx] = {}
                curr_input = self.input[idx]
                
                for feature in features:
                    curr_input[feature] = self.__read_feature(feature)
            
            for idx in range(output_lines):
                self.output[idx] = {}
                curr_output = self.output[idx]
                
                for feature in features: 
                    curr_output[feature] = self.__read_feature(feature)
            
            for idx in range(bidirectional_lines):
                self.bidirectional[idx] = {}
                curr_bidir = self.bidirectional[idx]

                for feature in features:
                    curr_bidir[feature] = self.__read_feature(feature)

            if i3c_bank:
                count = self.__read_val(2, byteorder="little")
                self.i3c_bank = self.__read_val_array(word_size=2, num_words=count, byteorder="little")
            
            end_align = self.__read_val(word_size=2, byteorder="little")
            assert end_align==1, "End Alignment incorrect, parsing failed"
                

    def pioStringToBin (self, string_val:str) -> int:
        return self.str_pio_mappings.get(string_val, 0)

    def pioBinToString (self, binary_val:int) -> str:
        return self.pio_str_mappings.get(binary_val, "UNKNOWN")

    def csv_repr(self, options):
        if isinstance(options, int):
            return self.pioBinToString(options)
        else:
            return "/".join(self.pioBinToString(option) for option in options)

    def export_csv(self, csv_file:str, section:str=None, force=False):
        section = section.lower() 
        if (section == "input"):
            data = self.input
        elif section == "output":
            data = self.output
        elif section == "bidirectional":
            data = self.bidirectional
        else:
            raise Exception(f"Unknown INI section {section} requested for export")

        try:
            csv_file = Path(csv_file)
        except:
            raise Exception("Invalid epxort path")
        if csv_file.exists() and not force:
            raise FileExistsError (f"file {csv_file} already exists, call with `force=True` to overwrite")
        else:
            with open(csv_file, "w") as f:
                writer = csv.writer(f)
                features = [feature.value for feature in self.schema["features"]]
                row_data = [""]
                row_data.extend(feature["name"] for feature in features)
                writer.writerow(row_data)
                
                for idx in range(self.schema[section]):
                    data_dict = data.get(idx, {})
                    row_data = [idx+1]
                    row_data.extend([self.csv_repr(data_dict.get(feature, "")) for feature in self.schema["features"]])
                    writer.writerow(row_data)
    
    def __read_val(self, word_size=2, byteorder='little'):
        pointer_loc, value = IniParser.read_val(byte_array = self.byte_array, 
                                                pointer_start=self.__pointer, 
                                                word_size=word_size,
                                                byteorder = byteorder)
        self.__pointer = pointer_loc
        return value
    
    def __read_val_array(self, word_size=2, num_words=1, byteorder='little'):
        pointer_loc, option_list = IniParser.read_val_array(self.byte_array, self.__pointer, word_size=word_size, num_words=num_words, byteorder=byteorder)
        self.__pointer = pointer_loc
        return option_list
    

    def __read_feature(self, feature:IOF):
        feature = feature.value
        count_size = feature["count"]
        feature_size = feature["size"]

        if count_size >= 0:
            count = self.__read_val(count_size, byteorder="little")
            feature_options = self.__read_val_array(feature_size, count, byteorder="little")
            return feature_options
        else:
            feature_val = self.__read_val(feature_size, byteorder="little")
            return feature_val
    

    @staticmethod
    def read_val(byte_array: bytes, pointer_start:int=0, word_size:int=1, byteorder='little') -> tuple[int, int]:
        pointer_end_pos = pointer_start + word_size
        read_bytes = byte_array[pointer_start: pointer_end_pos]
        val = int.from_bytes(read_bytes, byteorder)
        # unique_vals.add(val)
        return pointer_end_pos, val

    @staticmethod
    def read_val_array(byte_array:bytes, pointer_start:int=0, word_size:int=1, num_words:int=1, byteorder='little') -> tuple[int, list[int]]:
        val_list = [0] * num_words
        pointer_loc = pointer_start
        for idx in range(num_words):
            pointer_loc, val_list[idx] = IniParser.read_val(byte_array, pointer_loc, word_size, byteorder)
        return pointer_loc, val_list




if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="ini_h4x",
        description="This programme parses a .ini file and exports it as a csv file"
    )

    parser.add_argument('device', help="Device Name")
    parser.add_argument("-n", "--inifile", required=False, help="Manually set .ini file. ini_h4x will attempt to find the file using the device name otherwise")
    parser.add_argument('-s', '--section', default="all", choices=["input", "output", "bidirectional", "all"], required=False, help="Section of ini file to parse, default is 'all' ")
    parser.add_argument('-e', '--export_file', required=False, help="Name to use for output file")
    parser.add_argument("-f", '--force', required=False, help="set to overwrite conflicting existing files", action="store_true")


    args = parser.parse_args()
    device=args.device
    inifile = Path(args.inifile) if args.inifile else None
    section = args.section

    iniParser = IniParser(device=args.device.upper(), ini_file=inifile)
    iniParser.parse()

    sections = []
    if section == "all":
        sections = ["input", "output", "bidirectional"]
    else:
        sections = [section]
    
    for curr_section in sections: 
        if not args.export_file:
            export_file = f"{device}.{curr_section}.csv"
            export_file = Path(export_file)
        else:
            export_file = args.export_file
            export_file = export_file[:-4] if export_file[-3:].lower()==".csv" else export_file

            if args.section == "all":
                export_file = f"{export_file}.{curr_section}.csv"
            else: 
                export_file = Path(f"{export_file}.csv")

        try:
            iniParser.export_csv(section=curr_section, csv_file=export_file, force=args.force)
        except FileExistsError:
            print("Export FAILED because export file already exists. Call with '-f' to overwrite existing file")


import sys
from bslib import read_bitstream
from PIL import Image

arr = read_bitstream(sys.argv[1])
if len(sys.argv) > 2:
    diff = read_bitstream(sys.argv[2])
    arr ^= diff

size = (arr.shape[1]*8, arr.shape[0])
print(size)
im = Image.frombytes(mode='1', size=size, data=arr)
#im.show()
im.save("bitmap.png","PNG")



from math import ceil
import array
from apycula import bitmatrix
from apycula.crc16 import make_crc16_arc

_B2B = [f"{i:08b}" for i in range(256)]

def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]

def bytearr(frame):
    "array of all bytes of the frame"
    return bytearray([int(n, base=2) for n in chunks(frame.strip(), 8)])

def bitarr(frame, pad):
    "Array of *content* bits"
    data = frame.strip()[pad:-64]
    return [int(n, base=2) for n in data]

def read_bitstream_version(fname):
    ver = "UNKNOWN"
    with open(fname) as inp:
        for line in inp:
            if line.startswith("//Tool Version:"):
                ver = line[16:]
                break
    return ver

def read_bitstream(fname):
    bitmap = []
    returnBitmap = []
    hdr = []
    ftr = []
    # additional slots
    # { slot_no: bitmap }
    extra_slots = {}
    current_slot = 0xff
    is_hdr = True
    crcdat = bytearray()
    preamble = 3
    frames = 0
    c = 0
    is5ASeries = False

    calc = make_crc16_arc()
    compressed = False
    compress_keys = {}
    with open(fname) as inp:
        for line in inp:
            if line.startswith("//"):
                #print("line: ", line)
                continue
            ba = bytearr(line)
            if not frames:
                if is_hdr:
                    #print("header:", ba)
                    hdr.append(ba)
                    if ba[0] == 0x10 and (int.from_bytes(ba, 'big') & (1 << 13)):
                        compressed = True
                    if ba[0] == 0x51:
                        compress_keys[f'{ba[5]:08b}'] = 8
                        if ba[6]:
                            compress_keys[f'{ba[6]:08b}'] = 4
                            if ba[7]:
                                compress_keys[f'{ba[7]:08b}'] = 2
                else:
                    #print("footer:", ba)
                    # Slots
                    if ba[0] == 0x6a:
                        if ba[7] == 0xff: # start of slots
                            continue
                        current_slot = ba[7]
                        #print("Slot:", current_slot)
                        continue
                    if ba[0] == 0x6b and ba[1] == 0x80 and ba[2] == 0: # slot data
                        slot_size = ba[3]
                        #print("Slot size:", slot_size)
                        slot_bitmap = []
                        for byt in chunks(line.strip()[8 * 4:-8 * 18], 8):
                            slot_bitmap.append([int(n, base=2) for n in byt])
                        slot_bitmap = bitmatrix.fliplr(slot_bitmap)
                        extra_slots[current_slot] = bitmatrix.transpose(slot_bitmap)
                        #for rd in extra_slots[current_slot]:
                        #    print(rd)
                        continue
                    ftr.append(ba)
                if not preamble and ba[0] != 0xd2: # SPI address
                    #print("spi address", ba)
                    crcdat.extend(ba)
                if not preamble and ba[0] == 0x3b: # frame count
                    frames = int.from_bytes(ba[2:], 'big')
                    #print(f"frames:{frames}");
                    is_hdr = False
                if not preamble and ba[0] == 0x06: # device ID
                    if ba == b'\x06\x00\x00\x00\x11\x00\x58\x1b':     # GW1N-9
                        padding = 4
                        compress_padding = 44 # <-- when using compression, the width of one row in bits must be a multiple of 64
                    elif ba == b'\x06\x00\x00\x00\x11\x00H\x1b':      # GW1N-9C
                        padding = 4
                        compress_padding = 44
                    elif ba == b'\x06\x00\x00\x00\x09\x00\x28\x1b':   # GW1N-1
                        padding = 0
                        compress_padding = 0
                    elif ba == b'\x06\x00\x00\x00\x01\x008\x1b':      # GW1N-4
                        padding = 0
                        compress_padding = 8
                    elif ba == b'\x06\x00\x00\x00\x01\x00h\x1b':      # GW1NZ-1
                        padding = 0
                        compress_padding = 0
                    elif ba == b'\x06\x00\x00\x00\x01\x00\x98\x1b':   # GW1NS-4
                        padding = 0
                        compress_padding = 8
                    elif ba == b'\x06\x00\x00\x00\x00\x00\x08\x1b':   # GW2A-18(C)
                        padding = 0
                        compress_padding = 16
                    elif ba == b'\x06\x00\x00\x00\x00\x01\x28\x1b':   # GW5A-25A
                        padding = 3
                        compress_padding = 43
                        is5ASeries = True
                    elif ba == b'\x06\x00\x00\x00\x00\x01\x08\x1b':   # GW5AST-138C
                        padding = 3
                        compress_padding = 43
                        is5ASeries = True
                    else:
                        raise ValueError("Unsupported device", ba)
                preamble = max(0, preamble-1)
                continue
            if is5ASeries == False:
                crcdat.extend(ba[:-8])
                crc1 = (ba[-7] << 8) + ba[-8]
                crc2 = calc(crcdat)
                assert crc1 == crc2, f"Not equal {crc1} {crc2} for {crcdat}"
                if crc1 != crc2:
                    print("frame: ", c, ba, len(ba))
                    print("crcdata: ", crcdat, len(crcdat))
                    print("crc error - frame:", c, frames, " : ", crc1, " != ", crc2)

                crcdat = ba[-6:]
            if compressed:
                uncompressed_line = ''
                for byte_str in chunks(line[:-64], 8):
                    if byte_str in compress_keys:
                        for _ in range(compress_keys[byte_str]):
                            uncompressed_line += "00000000"
                    else:
                        uncompressed_line += byte_str

                uncompressed_line += line[-64:]
                bitmap.append(bitarr(uncompressed_line, compress_padding))
            else:
                bitmap.append(bitarr(line, padding))

            frames = max(0, frames-1)
            c = c + 1

        returnBitmap = bitmatrix.fliplr(bitmap)
        if is5ASeries:
            returnBitmap = bitmatrix.transpose(returnBitmap)

        return returnBitmap, hdr, ftr, extra_slots

def compressLine(line, key8Z, key4Z, key2Z):
    newline = []
    for i in range(0, len(line), 8):
        val = array.array('B', line[i:i+8]).tobytes().replace(8 * b'\x00', bytes([key8Z]))
        if key4Z:
            val = val.replace(4 * b'\x00', bytes([key4Z]))
            if key2Z:
                val = val.replace(2 * b'\x00', bytes([key2Z]))
        newline += val
    return newline

def write_gw5_bsram_init_map(f, crcdat, calc, gw5a_bsram_init_map, gw5a_bsrams):
    # BSRAM init part. Count used columns
    last_col = -1
    used_blocks = 0
    block_seq = {}   # start col of block sequence: number of blocks
    last_block_seq = None
    for bsram in gw5a_bsrams:
        col, _, _, _, _ = bsram
        if col != last_col:
            used_blocks += 1
            if col - last_col != 3:         # 3 cells for one BSRAM, if the columns are 3 cells apart
                block_seq.setdefault(col // 3, 0)  #  when the BSRAMs are neighbours
                last_block_seq = col // 3
            block_seq[last_block_seq] += 1
            last_col = col
    # rearrange init map - cut unused tail blocks - we don't need zero init data after last used block
    tail = used_blocks * 256
    w = bitmatrix.shape(gw5a_bsram_init_map)[1]
    bitInitMap = [[gw5a_bsram_init_map[i][w - j - 1] for j in range(w)] for i in range(tail)]
    assert bitmatrix.shape(bitInitMap)[1] % 8 == 0
    byteInitMap = bitmatrix.packbits(bitInitMap, axis = 1)

    # write BSRAM init data
    data_first_col = 0
    for start, cnt in block_seq.items():
        ba = bytearray(b'\x12\x00\x00\x00')
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        f.write('\n')

        # empty cols
        ba = bytearray(b'\x70\x00\x00')
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        ba = bytearray.fromhex(f"{start + 1:02x}")
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        ba = bytearray(b'\x00' * (start + 1))
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        f.write('\n')

        # data cols
        ba = bytearray(b'\x4e\x80')
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        ba = bytearray.fromhex(f"{cnt % 256 :02x}")
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        ba = bytearray.fromhex(f"{cnt >> 8 :02x}")
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        f.write('\n')

        # data
        for data_row in byteInitMap[data_first_col : data_first_col + 256 * cnt]:
            f.write(''.join(_B2B[b] for b in data_row))
            crcdat.extend(data_row)
            crc_ = calc(crcdat)
            crcdat = bytearray(b'\xff'*6)
            f.write(_B2B[crc_&0xff]+_B2B[crc_>>8])
            f.write('1'*48)
            f.write('\n')

        data_first_col += 256 * cnt

        # end of block
        ba = bytearray(b'\xff' * 18)
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        crc_ = calc(crcdat)
        crcdat = bytearray()
        f.write(_B2B[crc_&0xff]+_B2B[crc_>>8])
        f.write('\n')

def write_gw5_138_bsram_init_map(f, crcdat, calc, gw5a_bsram_init_map, gw5a_bsrams):
    # BSRAM init part. Count used columns
    last_col = -1
    used_blocks = 0
    block_seq = {}   # start col of block sequence: number of blocks
    last_block_seq = None
    for bsram in gw5a_bsrams:
        col, _, _, _, _ = bsram
        if col != last_col:
            used_blocks += 1
            if col - last_col != 3:         # 3 cells for one BSRAM, if the columns are 3 cells apart
                block_seq.setdefault(col // 3, 0)  #  when the BSRAMs are neighbours
                last_block_seq = col // 3
            block_seq[last_block_seq] += 1
            last_col = col
    # rearrange init map - cut unused tail blocks - we don't need zero init data after last used block
    tail = used_blocks * 256
    w = bitmatrix.shape(gw5a_bsram_init_map)[1]
    bitInitMap = [[gw5a_bsram_init_map[i][w - j - 1] for j in range(w)] for i in range(tail)]
    assert bitmatrix.shape(bitInitMap)[1] % 8 == 0
    byteInitMap = bitmatrix.packbits(bitInitMap, axis = 1)

    # write BSRAM init data
    data_first_col = 0

    ba = bytearray(b'\x12\x00\x00\x00')
    crcdat.extend(ba)
    f.write(''.join(_B2B[b] for b in ba))
    f.write('\n')

    for start, cnt in block_seq.items():
        # write address if col != 0

        if start != 0:
            ba = bytearray(2738)
            ba[0] = 0x98
            ba[2720 - 45 * start] = (1 << 6) # onehot address: one bit set for first col
            f.write(''.join(_B2B[b] for b in ba))
            f.write('\n')
             # 0x98 command is *not* included in CRC calculation

        # data cols
        ba = bytearray(b'\x4e\x80')
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        ba = bytearray.fromhex(f"{cnt:02x}")
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        ba = bytearray(b'\x01' if start == 0 else b'\x00') # 257 in first col only
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        f.write('\n')

        # extra zero frame for first col
        if start == 0:
            extra_row = bytearray(len(byteInitMap[data_first_col]))
            f.write(''.join(_B2B[b] for b in extra_row))
            crcdat.extend(extra_row)
            crc_ = calc(crcdat)
            crcdat = bytearray(b'\xff'*6)
            f.write(_B2B[crc_&0xff]+_B2B[crc_>>8])
            f.write('1'*48)
            f.write('\n')

        # data
        for data_row in byteInitMap[data_first_col : data_first_col + 256 * cnt]:
            f.write(''.join(_B2B[b] for b in data_row))
            crcdat.extend(data_row)
            crc_ = calc(crcdat)
            crcdat = bytearray(b'\xff'*6)
            f.write(_B2B[crc_&0xff]+_B2B[crc_>>8])
            f.write('1'*48)
            f.write('\n')

        data_first_col += 256 * cnt

        # end of block
        ba = bytearray(b'\xff' * 18)
        crcdat.extend(ba)
        f.write(''.join(_B2B[b] for b in ba))
        crc_ = calc(crcdat)
        crcdat = bytearray()
        f.write(_B2B[crc_&0xff]+_B2B[crc_>>8])
        f.write('\n')

def write_bitstream_with_bsram_init(fname, bs, hdr, ftr, compress, extra_slots, bsram_init):
    new_bs = bitmatrix.vstack(bs, bsram_init)
    new_hdr = hdr.copy()
    write_bitstream(fname, new_bs, new_hdr, ftr, compress, extra_slots)

def write_bitstream(fname, bs, hdr, ftr, compress, extra_slots, gw5a_bsram_init_map = None, gw5a_bsrams = None, is_gw5a_138 = False):
    bs = bitmatrix.fliplr(bs)
    hdr[-1][2:] = int(bitmatrix.shape(bs)[0]).to_bytes(2, 'big')

    if compress:
        padlen = (ceil(bitmatrix.shape(bs)[1] / 64) * 64) - bitmatrix.shape(bs)[1]
    else:
        padlen = (ceil(bitmatrix.shape(bs)[1] / 8) * 8) - bitmatrix.shape(bs)[1]
    pad = bitmatrix.ones(bitmatrix.shape(bs)[0], padlen)
    no_compress_pad_bytes = (padlen - ((ceil(bitmatrix.shape(bs)[1] / 8) * 8) - bitmatrix.shape(bs)[1])) // 8
    bs = bitmatrix.hstack(pad, bs)
    assert bitmatrix.shape(bs)[1] % 8 == 0
    bs = bitmatrix.packbits(bs, axis = 1)

    unused_bytes = []
    if compress:
        # search for smallest values not used in the bitstream
        lst = bitmatrix.byte_histogram(bs)
        unused_bytes = [i for i,val in enumerate(lst) if val==0]
        if unused_bytes:
            # We may simply not have the bytes we need for the keys.
            [key8Z, key4Z, key2Z] = (unused_bytes + [0, 0])[0:3]
            # update line 0x10 with compress enable bit
            hdr10 = int.from_bytes(hdr[4], 'big') | (1 << 13)
            hdr[4] = bytearray.fromhex(f"{hdr10:016x}")

            # update line 0x51 with keys
            hdr51 = int.from_bytes(hdr[5], 'big') & ~0xffffff
            hdr51 = hdr51 | (key8Z << 16) | (key4Z << 8) | (key2Z)
            hdr[5] = bytearray.fromhex(f"{hdr51:016x}")
        else:
            print("Warning. No unused bytes, will be uncompressed.")

    crcdat = bytearray()
    preamble = 3
    calc = make_crc16_arc()
    with open(fname, 'w') as f:
        for ba in hdr:
            if not preamble and ba[0] != 0xd2: # SPI address
                crcdat.extend(ba)
            preamble = max(0, preamble-1)
            f.write(''.join(_B2B[b] for b in ba))
            f.write('\n')
        for ba in bs:
            if compress:
                if unused_bytes:
                    ba = compressLine(ba, key8Z, key4Z, key2Z)
                else:
                    ba = ba[no_compress_pad_bytes : ]
            f.write(''.join(_B2B[b] for b in ba))
            crcdat.extend(ba)
            crc_ = calc(crcdat)
            crcdat = bytearray(b'\xff'*6)
            f.write(_B2B[crc_&0xff]+_B2B[crc_>>8])
            f.write('1'*48)
            f.write('\n')

        # end of main grid
        f.write(''.join(_B2B[b] for b in ftr[0]))
        f.write('\n')

        if extra_slots:
            # slot preamble
            crcdat = bytearray()
            ba = bytearray(b'\x6a\x00\x00\x00\x00\x00\x00\xff')
            crcdat.extend(ba)
            f.write(''.join(_B2B[b] for b in ba))
            f.write('\n')
            ba = bytearray(b'\x6d\x00\x00\x00')
            crcdat.extend(ba)
            f.write(''.join(_B2B[b] for b in ba))
            ba = bytearray(b'\xff'*16)
            f.write(''.join(_B2B[b] for b in ba))
            f.write('\n')

            for slot_idx, slot_bitmap in extra_slots.items():
                # slot header
                ba = bytearray(b'\x6a\x00\x00\x00\x00\x00\x00')
                crcdat.extend(ba)
                f.write(''.join(_B2B[b] for b in ba))
                ba = bytearray.fromhex(f"{slot_idx:02x}")
                crcdat.extend(ba)
                f.write(''.join(_B2B[b] for b in ba))
                f.write('\n')

                ba = bytearray(b'\x6b\x80\x00')
                crcdat.extend(ba)
                f.write(''.join(_B2B[b] for b in ba))
                shape = bitmatrix.shape(slot_bitmap)
                ba = bytearray.fromhex(f"{shape[0] * shape[1] // 8:02x}")
                crcdat.extend(ba)
                f.write(''.join(_B2B[b] for b in ba))

                # slot bitmap
                bs = bitmatrix.transpose(slot_bitmap)
                bs = bitmatrix.fliplr(bs)
                bs = bitmatrix.packbits(bs, axis = 1)
                for ba in bs:
                    crcdat.extend(ba)
                    f.write(''.join(_B2B[b] for b in ba))
                crc_ = calc(crcdat)
                crcdat = bytearray(b'\xff'*2)
                f.write(_B2B[crc_&0xff]+_B2B[crc_>>8])
                f.write('1'*128)
                f.write('\n')
        else:
            crcdat = bytearray()
        if gw5a_bsram_init_map is not None:
            if is_gw5a_138:
                write_gw5_138_bsram_init_map(f, crcdat, calc, gw5a_bsram_init_map, gw5a_bsrams)
                crcdat = bytearray()
            else:
                write_gw5_bsram_init_map(f, crcdat, calc, gw5a_bsram_init_map, gw5a_bsrams)
                crcdat = bytearray()

        for ba in ftr[1:]:
            f.write(''.join(_B2B[b] for b in ba))
            f.write('\n')


def display(fname, data):
    from PIL import Image
    """
    im = Image.frombytes(
            mode='1',
            size=data.shape[::-1],
            data=bitmatrix.packbits(data, axis = 1))
    """

    tdata = bitmatrix.packbits(data, axis = 1)
    im = Image.new('RGB', bitmatrix.shape(tdata)[::-1], 255)
    idata = im.load()
    for x in range(im.size[0]):
        for y in range(im.size[1]):
            idata[x, y] = (tdata[y][x], tdata[y][x], tdata[y][x])
    if fname:
        im.save(fname)
    return im
